# Unit test check_residuals_input
# ==============================================================================
import re
import pytest
import numpy as np
from skforecast.utils import check_residuals_input
from skforecast.exceptions import UnknownLevelWarning


@pytest.mark.parametrize("residuals", 
                         [None, {}, np.array([])],
                         ids = lambda res: f'residuals: {res}')
@pytest.mark.parametrize("use_binned_residuals", 
                         [True, False],
                         ids = lambda binned: f'use_binned_residuals: {binned}')
def test_check_residuals_input_ValueError_when_not_in_sample_residuals(residuals, use_binned_residuals):
    """
    Test ValueError is raised when there is no in_sample_residuals_ or 
    in_sample_residuals_by_bin_.
    """

    if use_binned_residuals:
        literal = "in_sample_residuals_by_bin_"
    else:
        literal = "in_sample_residuals_"

    err_msg = re.escape(
        f"`forecaster.{literal}` is either None or empty. Use "
        f"`store_in_sample_residuals = True` when fitting the forecaster "
        f"or use the `set_in_sample_residuals()` method before predicting."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_residuals_input(
            forecaster_name              = 'ForecasterRecursive',
            use_in_sample_residuals      = True,
            in_sample_residuals_         = residuals,
            out_sample_residuals_        = None,
            use_binned_residuals         = use_binned_residuals,
            in_sample_residuals_by_bin_  = residuals,
            out_sample_residuals_by_bin_ = None
        )


@pytest.mark.parametrize("residuals", 
                         [None, {}, np.array([])],
                         ids = lambda res: f'residuals: {res}')
@pytest.mark.parametrize("use_binned_residuals", 
                         [True, False],
                         ids = lambda binned: f'use_binned_residuals: {binned}')
def test_check_residuals_input_ValueError_when_not_out_sample_residuals(residuals, use_binned_residuals):
    """
    Test ValueError is raised when there is no out_sample_residuals_ or 
    out_sample_residuals_by_bin_.
    """

    if use_binned_residuals:
        literal = "out_sample_residuals_by_bin_"
    else:
        literal = "out_sample_residuals_"

    err_msg = re.escape(
        f"`forecaster.{literal}` is either None or empty. Use "
        f"`use_in_sample_residuals = True` or the "
        f"`set_out_sample_residuals()` method before predicting."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_residuals_input(
            forecaster_name              = 'ForecasterRecursive',
            use_in_sample_residuals      = False,
            in_sample_residuals_         = None,
            out_sample_residuals_        = residuals,
            use_binned_residuals         = use_binned_residuals,
            in_sample_residuals_by_bin_  = None,
            out_sample_residuals_by_bin_ = residuals
        )


@pytest.mark.parametrize("use_binned_residuals", 
                         [True, False],
                         ids = lambda binned: f'use_binned_residuals: {binned}')
def test_check_residuals_input_multiseries_ValueError_when_not_in_sample_residuals_for_any_level(use_binned_residuals):
    """
    Test ValueError is raised when there is no in_sample_residuals_ for any level
    in 'ForecasterRecursiveMultiSeries'.
    """
    levels = ['1', '2']
    residuals = {'1': np.array([1, 2, 3, 4, 5])}

    if use_binned_residuals:
        residuals = {
            '1': {1: np.array([1, 2, 3, 4, 5])},
            '_unknown_level': {1: np.array([1, 2, 3, 4, 5])}
        }
        literal = "in_sample_residuals_by_bin_"
    else:
        residuals = {
            '1': np.array([1, 2, 3, 4, 5]),
            '_unknown_level': np.array([1, 2, 3, 4, 5])
        }
        literal = "in_sample_residuals_"

    warn_msg = re.escape(
        f"`levels` {set('2')} are not present in `forecaster.{literal}`, "
        f"most likely because they were not present in the training data. "
        f"A random sample of the residuals from other levels will be used. "
        f"This can lead to inaccurate intervals for the unknown levels."
    )
    with pytest.warns(UnknownLevelWarning, match = warn_msg):
        check_residuals_input(
            forecaster_name              = 'ForecasterRecursiveMultiSeries',
            levels                       = levels,
            encoding                     ='ordinal',
            use_in_sample_residuals      = True,
            in_sample_residuals_         = residuals,
            out_sample_residuals_        = None,
            use_binned_residuals         = use_binned_residuals,
            in_sample_residuals_by_bin_  = residuals,
            out_sample_residuals_by_bin_ = None,
        )


@pytest.mark.parametrize("use_binned_residuals", 
                         [True, False],
                         ids = lambda binned: f'use_binned_residuals: {binned}')
def test_check_residuals_input_multiseries_ValueError_when_not_out_sample_residuals_for_any_level(use_binned_residuals):
    """
    Test ValueError is raised when there is no out_sample_residuals_ for any level
    in 'ForecasterRecursiveMultiSeries'.
    """
    levels = ['1', '2']
    residuals = {'1': np.array([1, 2, 3, 4, 5])}

    if use_binned_residuals:
        residuals = {
            '1': {1: np.array([1, 2, 3, 4, 5])},
            '_unknown_level': {1: np.array([1, 2, 3, 4, 5])}
        }
        literal = "out_sample_residuals_by_bin_"
    else:
        residuals = {
            '1': np.array([1, 2, 3, 4, 5]),
            '_unknown_level': np.array([1, 2, 3, 4, 5])
        }
        literal = "out_sample_residuals_"

    warn_msg = re.escape(
        f"`levels` {set('2')} are not present in `forecaster.{literal}`. "
        f"A random sample of the residuals from other levels will be used. "
        f"This can lead to inaccurate intervals for the unknown levels. "
        f"Otherwise, Use the `set_out_sample_residuals()` method before "
        f"predicting to set the residuals for these levels.",
    )
    with pytest.warns(UnknownLevelWarning, match = warn_msg):
        check_residuals_input(
            forecaster_name              = 'ForecasterRecursiveMultiSeries',
            levels                       = levels,
            encoding                     ='ordinal',
            use_in_sample_residuals      = False,
            in_sample_residuals_         = None,
            out_sample_residuals_        = residuals,
            use_binned_residuals         = use_binned_residuals,
            in_sample_residuals_by_bin_  = None,
            out_sample_residuals_by_bin_ = residuals,
        )


@pytest.mark.parametrize("forecaster_name", 
                         ['ForecasterRecursiveMultiSeries', 'ForecasterDirectMultiVariate', 'ForecasterRnn'],
                         ids = lambda fn: f'forecaster_name: {fn}')
@pytest.mark.parametrize("use_binned_residuals", 
                         [True, False],
                         ids = lambda binned: f'use_binned_residuals: {binned}')
def test_check_residuals_input_ValueError_when_residuals_for_some_level_is_None(forecaster_name, use_binned_residuals):
    """
    Test ValueError is raised when residuals for some level is None or empty
    in 'ForecasterRecursiveMultiSeries'.
    """
    levels = ['1', '2']

    if use_binned_residuals:
        residuals = {
            'l1': {1: np.array([1, 2, 3, 4, 5])},
            'l2': {2: np.array([1, 2, 3, 4, 5])},
            'l3': {},
            '_unknown_level': {1: np.array([1, 2, 3, 4, 5])}
        }
        use_in_sample_residuals = False
        literal = "out_sample_residuals_by_bin_"
    else:
        residuals = {
            'l1': np.array([1, 2, 3, 4, 5]),
            'l2': np.array([1, 2, 3, 4, 5]),
            'l3': None,
            '_unknown_level': np.array([1, 2, 3, 4, 5])
        }
        use_in_sample_residuals = True
        literal = "in_sample_residuals_"

    err_msg = re.escape(
        f"Residuals for level 'l3' are None. Check `forecaster.{literal}`."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_residuals_input(
            forecaster_name              = forecaster_name,
            levels                       = levels,
            encoding                     ='ordinal',
            use_in_sample_residuals      = use_in_sample_residuals,
            in_sample_residuals_         = residuals,
            out_sample_residuals_        = residuals,
            use_binned_residuals         = use_binned_residuals,
            in_sample_residuals_by_bin_  = residuals,
            out_sample_residuals_by_bin_ = residuals
        )



# Unit test _backtesting_forecaster No refit
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from scipy.stats import norm
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from skforecast.recursive import ForecasterRecursive
from skforecast.direct import ForecasterDirect
from skforecast.model_selection._validation import _backtesting_forecaster
from skforecast.model_selection._split import TimeSeriesFold

# Fixtures
from ..fixtures_model_selection import y
from ..fixtures_model_selection import exog
from ..fixtures_model_selection import out_sample_residuals


# ******************************************************************************
# * Test _backtesting_forecaster No Interval                                   *
# ******************************************************************************

@pytest.mark.parametrize("n_jobs", [-1, 1, 'auto'],
                         ids=lambda n: f'n_jobs: {n}')
def test_output_backtesting_forecaster_no_exog_no_remainder_ForecasterRecursive_with_mocked(n_jobs):
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    ForecasterRecursive.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.0646438286283131]})
    expected_predictions = pd.DataFrame({
        'pred':np.array([0.55717779, 0.43355138, 0.54969767, 0.52945466, 
                         0.39585199, 0.55935949, 0.45263533, 0.4578669 , 
                         0.36988237, 0.57912951, 0.48686057, 0.45709952])}, 
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )

    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]

    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster = forecaster,
                                       y          = y,
                                       cv         = cv,
                                       exog       = None,
                                       metric     = 'mean_squared_error',
                                       n_jobs     = n_jobs,
                                       verbose    = False
                                   )
                                   
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_no_exog_no_remainder_ForecasterDirect_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    ForecasterDirect.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.06999891546733726]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.5468482,
                    0.44670961,
                    0.57651222,
                    0.52511275,
                    0.39745044,
                    0.53365885,
                    0.39390143,
                    0.47002119,
                    0.38080838,
                    0.56987255,
                    0.42302249,
                    0.45580163,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    forecaster = ForecasterDirect(
                     regressor = LinearRegression(), 
                     lags      = 3,
                     steps     = 4
                 )
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = None,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )
                                   
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_no_exog_no_initial_train_size_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    no initial_train_size, steps=1, ForecasterRecursive.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.05194702533929101]})
    expected_predictions = pd.DataFrame({
        'pred': np.array([0.50394528, 0.53092847, 0.51638165, 0.47382814, 0.61996956,
                         0.47685471, 0.52565717, 0.50842469, 0.4925563 , 0.55717972,
                         0.45707228, 0.45990822, 0.53762873, 0.5230163 , 0.41661072,
                         0.51097738, 0.52448483, 0.48179537, 0.5307759 , 0.55580453,
                         0.51780297, 0.53189442, 0.55356883, 0.46142853, 0.52517734,
                         0.46241276, 0.49292214, 0.53169102, 0.40448875, 0.55686226,
                         0.46860633, 0.5098154 , 0.49041677, 0.48435035, 0.51152271,
                         0.56870534, 0.53226143, 0.49091506, 0.56878395, 0.42767269,
                         0.53335856, 0.48167273, 0.5658333 , 0.41464667, 0.56733702,
                         0.5724869 , 0.45299923])
        }, 
        index=pd.RangeIndex(start=3, stop=50, step=1)
    )
    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    forecaster.fit(y=y)
    
    cv = TimeSeriesFold(
            steps                 = 1,
            initial_train_size    = None,
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        cv         = cv,
                                        exog       = None,                                      
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )
                 
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_no_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.07085869503962372]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.55717779,
                    0.43355138,
                    0.54969767,
                    0.52945466,
                    0.48308861,
                    0.54252612,
                    0.46139434,
                    0.44730047,
                    0.50031862,
                    0.49831103,
                    0.55613172,
                    0.42970914,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )         
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_yes_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.05585411566592716]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.59059622,
                    0.47257504,
                    0.53024098,
                    0.46163343,
                    0.41316188,
                    0.51416101,
                    0.42705363,
                    0.4384041,
                    0.42611891,
                    0.59547291,
                    0.5170294,
                    0.4982889,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = exog,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_yes_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.06313056651237414]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.59059622,
                    0.47257504,
                    0.53024098,
                    0.46163343,
                    0.50035119,
                    0.49157332,
                    0.43327346,
                    0.42420804,
                    0.53389427,
                    0.51693094,
                    0.60207937,
                    0.48227974,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster         = forecaster,
                                        y                  = y,
                                        exog               = exog,
                                        cv                 = cv,
                                        metric             = 'mean_squared_error',
                                        verbose            = False
                                   )
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_yes_exog_yes_remainder_skip_folds_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked,
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.044538146622722964]})
    expected_predictions = pd.DataFrame(
        {
            "pred": [
                0.590596217638621,
                0.47257503863519656,
                0.5302409766813725,
                0.46163343193846784,
                0.5003511853456513,
                0.6020793655316975,
                0.4822797443738659,
            ],
            "fold": [0, 0, 0, 0, 0, 1, 1],
            "lag_1": [
                0.89338916,
                0.590596217638621,
                0.47257503863519656,
                0.5302409766813725,
                0.46163343193846784,
                0.98555979,
                0.6020793655316975,
            ],
            "lag_2": [
                0.42635131,
                0.89338916,
                0.590596217638621,
                0.47257503863519656,
                0.5302409766813725,
                0.48303426,
                0.98555979,
            ],
            "lag_3": [
                0.31226122,
                0.42635131,
                0.89338916,
                0.590596217638621,
                0.47257503863519656,
                0.25045537,
                0.48303426,
            ],
            "exog": [
                0.30476807,
                0.39818568,
                0.70495883,
                0.99535848,
                0.35591487,
                0.2408559,
                0.34345601,
            ],
        },
        index=pd.Index([38, 39, 40, 41, 42, 48, 49], dtype="int64"),
    )

    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12

    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = 2,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
        forecaster        = forecaster,
        y                 = y,
        exog              = exog,
        cv                = cv,
        metric            = "mean_squared_error",
        return_predictors = True,
        verbose           = True,
    )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Test _backtesting_forecaster Interval                                      *
# ******************************************************************************

def test_output_backtesting_forecaster_interval_no_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes.
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.06464382862831312]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.55717779, 0.19882822, 0.87939731],
            [0.43355138, 0.08976463, 0.86652665],
            [0.54969767, 0.20849403, 0.98096724],
            [0.52945466, 0.16705061, 0.92717139],
            [0.39585199, 0.03750242, 0.71807151],
            [0.55935949, 0.21557274, 0.99233476],
            [0.45263533, 0.11143169, 0.8839049 ],
            [0.4578669 , 0.09546285, 0.85558363],
            [0.36988237, 0.01153279, 0.69210188],
            [0.57912951, 0.23534276, 1.01210478],
            [0.48686057, 0.14565693, 0.91813015],
            [0.45709952, 0.09469547, 0.85481624]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    
    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = None,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False,
                                       show_progress           = False
                                   )
                            
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_interval_no_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.07085869503962372]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.55717779, 0.19882822, 0.87939731],
            [0.43355138, 0.08976463, 0.86652665],
            [0.54969767, 0.20849403, 0.98096724],
            [0.52945466, 0.16705061, 0.92717139],
            [0.48308861, 0.12658912, 0.89058516],
            [0.54252612, 0.18417655, 0.86474563],
            [0.46139434, 0.11760759, 0.89436961],
            [0.44730047, 0.10609683, 0.87857004],
            [0.50031862, 0.13791457, 0.89803535],
            [0.49831103, 0.14181154, 0.90580758],
            [0.55613172, 0.19778215, 0.87835123],
            [0.42970914, 0.08592239, 0.86268441]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )

    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = None,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_interval_yes_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes.
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.05585411566592716]})
    expected_predictions = pd.DataFrame(
    data = np.array([
            [0.59059622, 0.24316567, 0.93342085],
            [0.47257504, 0.12485033, 0.87643834],
            [0.53024098, 0.18273747, 0.93078198],
            [0.46163343, 0.10256219, 0.83853371],
            [0.41316188, 0.06573133, 0.75598651],
            [0.51416101, 0.16643631, 0.91802432],
            [0.42705363, 0.07955012, 0.82759464],
            [0.4384041 , 0.07933286, 0.81530438],
            [0.42611891, 0.07868836, 0.76894354],
            [0.59547291, 0.24774821, 0.99933622],
            [0.5170294 , 0.16952589, 0.91757041],
            [0.4982889 , 0.13921766, 0.87518918]]
        ),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )

    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_interval_yes_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.06313056651237414]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.59059622, 0.24316567, 0.93342085],
            [0.47257504, 0.12485033, 0.87643834],
            [0.53024098, 0.18273747, 0.93078198],
            [0.46163343, 0.10256219, 0.83853371],
            [0.50035119, 0.14089436, 0.88160183],
            [0.49157332, 0.14414278, 0.83439795],
            [0.43327346, 0.08554876, 0.83713677],
            [0.42420804, 0.07670453, 0.82474905],
            [0.53389427, 0.17482303, 0.91079456],
            [0.51693094, 0.15747411, 0.89818158],
            [0.60207937, 0.25464882, 0.944904  ],
            [0.48227974, 0.13455504, 0.88614305]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    
    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = (5, 95),
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False,
                                       show_progress           = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


@pytest.mark.parametrize("initial_train_size", 
                         [len(y) - 20, "2022-01-30 00:00:00"],
                         ids=lambda init: f'initial_train_size: {init}')
def test_output_backtesting_forecaster_no_refit_interval_yes_exog_bootstrapping(initial_train_size):
    """
    Test output of _backtesting_forecaster with predicted intervals as bootstrapping.
    """
    y_with_index = y.copy()
    y_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_with_index = exog.copy()
    exog_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')

    expected_metric = pd.DataFrame({"mean_squared_error": [0.06392277343002713]})
    expected_predictions = pd.DataFrame(
        data = np.array([
                [0.55538506, 0.60367643, 0.88152611, 0.5634774 , 0.72025328,
                    0.2137087 , 0.6199332 , 0.3945439 , 0.52863287, 0.40661365,
                    0.52863287],
                [0.44124674, 0.30814369, 0.67552661, 0.59094798, 0.1530626 ,
                    0.47746399, 0.30400363, 0.5175231 , 0.16374181, 0.33787653,
                    0.09089052],
                [0.49752366, 0.49160155, 0.17901185, 0.3860609 , 0.20553126,
                    0.42645717, 0.16942951, 0.68746286, 0.47814313, 0.49592495,
                    0.48960277],
                [0.53822797, 0.54949064, 0.60649888, 0.46744448, 0.39636506,
                    0.57650316, 0.31502897, 0.28183223, 0.71848603, 0.59646825,
                    0.29032951],
                [0.4793552 , 0.58652515, 0.45968808, 0.35057928, 0.41631354,
                    0.56344941, 0.13469427, 0.41615485, 0.4767692 , 0.47616   ,
                    0.30475643],
                [0.4474213 , 0.49571267, 0.77356235, 0.45551363, 0.61228952,
                    0.10574494, 0.51196944, 0.28658014, 0.42066911, 0.29864989,
                    0.42066911],
                [0.51546009, 0.38235704, 0.74973996, 0.66516132, 0.22727595,
                    0.55167734, 0.37821697, 0.59173645, 0.23795516, 0.41208988,
                    0.16510387],
                [0.54066381, 0.53474169, 0.22215199, 0.42920105, 0.24867141,
                    0.46959732, 0.21256965, 0.73060301, 0.52128328, 0.53906509,
                    0.53274291],
                [0.51970475, 0.53096742, 0.58797566, 0.44892126, 0.37784183,
                    0.55797994, 0.29650575, 0.26330901, 0.69996281, 0.57794502,
                    0.27180629],
                [0.51420212, 0.62137208, 0.494535  , 0.38542621, 0.45116047,
                    0.59829634, 0.1695412 , 0.45100177, 0.51161613, 0.51100693,
                    0.33960336],
                [0.4699511 , 0.51824248, 0.79609215, 0.47804344, 0.63481932,
                    0.12827474, 0.53449924, 0.30910994, 0.44319891, 0.3211797 ,
                    0.44319891],
                [0.55077981, 0.41767676, 0.78505967, 0.70048104, 0.26259567,
                    0.58699706, 0.41353669, 0.62705617, 0.27327487, 0.4474096 ,
                    0.20042359],
                [0.54063198, 0.53470987, 0.22212017, 0.42916922, 0.24863958,
                    0.46956549, 0.21253783, 0.73057119, 0.52125146, 0.53903327,
                    0.53271109],
                [0.48476663, 0.49602931, 0.55303754, 0.41398315, 0.34290372,
                    0.52304182, 0.26156764, 0.2283709 , 0.66502469, 0.54300691,
                    0.23686818],
                [0.49651396, 0.60368391, 0.47684684, 0.36773805, 0.43347231,
                    0.58060818, 0.15185304, 0.43331361, 0.49392796, 0.49331876,
                    0.32191519],
                [0.59616217, 0.64445355, 0.92230322, 0.60425451, 0.76103039,
                    0.25448581, 0.66071031, 0.43532101, 0.56940998, 0.44739077,
                    0.56940998],
                [0.46166238, 0.32855932, 0.69594224, 0.61136361, 0.17347823,
                    0.49787962, 0.32441926, 0.53793873, 0.18415744, 0.35829216,
                    0.11130615],
                [0.55197041, 0.54604829, 0.23345859, 0.44050765, 0.25997801,
                    0.48090392, 0.22387625, 0.74190961, 0.53258988, 0.55037169,
                    0.54404951],
                [0.5694184 , 0.58068107, 0.63768931, 0.49863492, 0.42755549,
                    0.60769359, 0.34621941, 0.31302266, 0.74967646, 0.62765868,
                    0.32151994],
                [0.52086432, 0.62803427, 0.5011972 , 0.39208841, 0.45782267,
                    0.60495854, 0.1762034 , 0.45766397, 0.51827832, 0.51766912,
                    0.34626555]]),
        columns = ['pred'] + [f'pred_boot_{i}' for i in range(10)],
        index = pd.date_range(start='2022-01-31', periods=20, freq='D')
    )

    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    cv = TimeSeriesFold(
             steps              = 5,
             initial_train_size = initial_train_size,
             refit              = False
         )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y_with_index,
                                       exog                    = exog_with_index,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = 'bootstrapping',
                                       interval_method         = 'bootstrapping', 
                                       n_boot                  = 10,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_no_refit_interval_distribution_yes_exog():
    """
    Test output of _backtesting_forecaster with predicted intervals as distribution.
    """
    y_with_index = y.copy()
    y_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_with_index = exog.copy()
    exog_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')

    expected_metric = pd.DataFrame({"mean_absolute_error": [0.19221231167961977]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.55538506, 0.5265697 , 0.19122853],
            [0.44124674, 0.45485962, 0.21296327],
            [0.49752366, 0.48702575, 0.20938072],
            [0.53822797, 0.51715637, 0.21511474],
            [0.4793552 , 0.4814511 , 0.20866791],
            [0.4474213 , 0.41860594, 0.19122853],
            [0.51546009, 0.52907297, 0.21296327],
            [0.54066381, 0.5301659 , 0.20938072],
            [0.51970475, 0.49863315, 0.21511474],
            [0.51420212, 0.51629803, 0.20866791],
            [0.4699511 , 0.44113574, 0.19122853],
            [0.55077981, 0.56439269, 0.21296327],
            [0.54063198, 0.53013407, 0.20938072],
            [0.48476663, 0.46369504, 0.21511474],
            [0.49651396, 0.49860986, 0.20866791],
            [0.59616217, 0.56734681, 0.19122853],
            [0.46166238, 0.47527525, 0.21296327],
            [0.55197041, 0.5414725 , 0.20938072],
            [0.5694184 , 0.5483468 , 0.21511474],
            [0.52086432, 0.52296022, 0.20866791]]),
        columns = ['pred', 'loc', 'scale'],
        index = pd.date_range(start='2022-01-31', periods=20, freq='D')
    )

    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    cv = TimeSeriesFold(
             steps              = 5,
             initial_train_size = len(y_with_index) - 20,
             refit              = False
         )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y_with_index,
                                       exog                    = exog_with_index,
                                       cv                      = cv,
                                       metric                  = 'mean_absolute_error',
                                       interval                = norm,
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 250,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


@pytest.mark.parametrize("interval", 
                         [0.90, (5, 95)], 
                         ids = lambda value: f'interval: {value}')
def test_output_backtesting_forecaster_interval_conformal_and_binned_with_mocked(interval):
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=5 (2 remainder), conformal=True, binned=True.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.063130566512374]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.59059622, 0.2974502 , 0.88374223],
            [0.47257504, 0.22356317, 0.72158691],
            [0.53024098, 0.14339956, 0.91708239],
            [0.46163343, 0.21262156, 0.7106453 ],
            [0.50035119, 0.35604735, 0.64465502],
            [0.49157332, 0.34726949, 0.63587716],
            [0.43327346, 0.18718506, 0.67936186],
            [0.42420804, 0.32142927, 0.52698681],
            [0.53389427, 0.2579754 , 0.80981315],
            [0.51693094, 0.13008952, 0.90377235],
            [0.60207937, 0.30893335, 0.89522538],
            [0.48227974, 0.0046491 , 0.95991039]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    
    forecaster = ForecasterRecursive(
        regressor=LinearRegression(), lags=3, binner_kwargs={'n_bins': 10}
    )
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
             steps                 = 5,
             initial_train_size    = len(y_train),
             window_size           = None,
             differentiation       = None,
             refit                 = False,
             fixed_train_size      = True,
             gap                   = 0,
             skip_folds            = None,
             allow_incomplete_fold = True,
             return_all_indexes    = False,
         )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster              = forecaster,
                                        y                       = y,
                                        exog                    = exog,
                                        cv                      = cv,
                                        metric                  = 'mean_squared_error',
                                        interval                = interval,
                                        interval_method         = 'conformal',
                                        use_in_sample_residuals = True,
                                        use_binned_residuals    = True,
                                        random_state            = 123,
                                        verbose                 = False,
                                        show_progress           = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


@pytest.mark.parametrize("interval", 
                         [0.90, (5, 95)], 
                         ids = lambda value: f'interval: {value}')
def test_output_backtesting_forecaster_interval_conformal_and_binned_with_mocked_ForecasterDirect(interval):
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=5 (2 remainder), conformal=True, binned=True.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.061964730085838]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.55584775, 0.37659309, 0.73510242],
            [0.45718425, 0.08077419, 0.83359431],
            [0.57242372, 0.36697393, 0.77787352],
            [0.47236513, 0.10146645, 0.84326382],
            [0.50476389, 0.16013181, 0.84939597],
            [0.51445718, 0.1698251 , 0.85908926],
            [0.42759863, 0.14064966, 0.71454759],
            [0.42937664, 0.14242768, 0.71632561],
            [0.51632981, 0.17169773, 0.86096189],
            [0.57758034, 0.37213055, 0.78303014],
            [0.55914632, 0.37989165, 0.73840098],
            [0.46393165, 0.09303296, 0.83483033]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    
    forecaster = ForecasterDirect(
        regressor=LinearRegression(), steps=5, lags=3, binner_kwargs={'n_bins': 10}
    )
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
             steps                 = 5,
             initial_train_size    = len(y_train),
             window_size           = None,
             differentiation       = None,
             refit                 = False,
             fixed_train_size      = True,
             gap                   = 0,
             skip_folds            = None,
             allow_incomplete_fold = True,
             return_all_indexes    = False,
         )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster              = forecaster,
                                        y                       = y,
                                        exog                    = exog,
                                        cv                      = cv,
                                        metric                  = 'mean_squared_error',
                                        interval                = interval,
                                        interval_method         = 'conformal',
                                        use_in_sample_residuals = True,
                                        use_binned_residuals    = True,
                                        random_state            = 123,
                                        verbose                 = False,
                                        show_progress           = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Out sample residuals                                                       *
# ******************************************************************************

def test_output_backtesting_forecaster_interval_out_sample_residuals_no_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes.
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error',
    'use_in_sample_residuals = False'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.0646438286283131]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.55717779, 0.63654358, 1.5406994 ],
            [0.43355138, 0.5760887 , 1.53342317],
            [0.54969767, 0.68344933, 1.57809559],
            [0.52945466, 0.69548679, 1.63622589],
            [0.39585199, 0.47521778, 1.3793736 ],
            [0.55935949, 0.70189681, 1.65923128],
            [0.45263533, 0.58638699, 1.48103325],
            [0.4578669 , 0.62389903, 1.56463813],
            [0.36988237, 0.44924816, 1.35340398],
            [0.57912951, 0.72166684, 1.67900131],
            [0.48686057, 0.62061223, 1.51525849],
            [0.45709952, 0.62313165, 1.56387074]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )

    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)    
    forecaster.out_sample_residuals_ = out_sample_residuals
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = None,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = False,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Callable metric                                                           *
# ******************************************************************************

def my_metric(y_true, y_pred):  # pragma: no cover
    """
    Callable metric
    """
    metric = ((y_true - y_pred) / len(y_true)).mean()
    
    return metric


def test_callable_metric_backtesting_forecaster_no_exog_no_remainder_with_mocked():
    """
    Test callable metric in _backtesting_forecaster with backtesting mocked, interval no. 
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    """
    expected_metric = pd.DataFrame({"my_metric": [0.005603130564222017]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.55717779,
                    0.43355138,
                    0.54969767,
                    0.52945466,
                    0.39585199,
                    0.55935949,
                    0.45263533,
                    0.4578669,
                    0.36988237,
                    0.57912951,
                    0.48686057,
                    0.45709952,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster = forecaster,
                                       y          = y,
                                       exog       = None,
                                       cv         = cv,
                                       metric     = my_metric,
                                       verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_list_metrics_backtesting_forecaster_no_exog_no_remainder_with_mocked():
    """
    Test list of metrics in _backtesting_forecaster with backtesting mocked, interval no. 
    Regressor is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    """
    expected_metrics = pd.DataFrame(
        {
            "mean_absolute_error": [0.20796216],
            "mean_squared_error": [0.0646438286283131],
        }
    )
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.55717779,
                    0.43355138,
                    0.54969767,
                    0.52945466,
                    0.39585199,
                    0.55935949,
                    0.45263533,
                    0.4578669,
                    0.36988237,
                    0.57912951,
                    0.48686057,
                    0.45709952,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )

    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metrics, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = None,
                                        cv         = cv,
                                        metric     = ['mean_absolute_error', mean_squared_error],
                                        verbose    = False
                                    )

    pd.testing.assert_frame_equal(expected_metrics, metrics)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Gap                                                                        *
# ******************************************************************************


def test_output_backtesting_forecaster_interval_yes_exog_yes_remainder_gap_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    20 observations to backtest, steps=5 and gap=3, metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.07965887934114284]})
    expected_predictions = pd.DataFrame(
        {
            "pred": {
                33: 0.6502299892039047,
                34: 0.5236463733888286,
                35: 0.4680925560056311,
                36: 0.48759731640448023,
                37: 0.4717244484701478,
                38: 0.5758584470242223,
                39: 0.5359293308956417,
                40: 0.48995122056218404,
                41: 0.4527131747771053,
                42: 0.49519412870799573,
                43: 0.2716782430953111,
                44: 0.2781382241553876,
                45: 0.31569423787508544,
                46: 0.39832240179281636,
                47: 0.3482268061581242,
                48: 0.6209868551179476,
                49: 0.4719075668681217,
            },
            "lower_bound": {
                33: 0.2718315398459972,
                34: 0.19453326150156153,
                35: 0.12440677892063873,
                36: 0.12107577460096824,
                37: 0.14533367434351524,
                38: 0.19745999766631478,
                39: 0.2068162190083746,
                40: 0.14626544347719167,
                41: 0.08619163297359334,
                42: 0.1688033545813632,
                43: -0.10672020626259643,
                44: -0.05097488773187947,
                45: -0.027991539209906935,
                46: 0.031800859989304375,
                47: 0.021836032031491648,
                48: 0.2425884057600401,
                49: 0.14279445498085463,
            },
            "upper_bound": {
                33: 1.065039136744312,
                34: 0.9230439995748042,
                35: 0.7539505750052855,
                36: 0.8314505919311639,
                37: 0.7469755216091047,
                38: 0.9906675945646294,
                39: 0.9353269570816173,
                40: 0.7758092395618383,
                41: 0.796566450303789,
                42: 0.7704452018469528,
                43: 0.6864873906357182,
                44: 0.6775358503413632,
                45: 0.6015522568747398,
                46: 0.7421756773195001,
                47: 0.6234778792970812,
                48: 1.0357960026583548,
                49: 0.8713051930540974,
            },
        }
    )
    expected_predictions = pd.DataFrame(
        data = np.array(
            [[ 0.65022999,  0.32111688,  1.0092382 ],
            [ 0.52364637,  0.17891138,  0.88265459],
            [ 0.46809256,  0.14156566,  0.82710077],
            [ 0.48759732,  0.16107043,  0.84202181],
            [ 0.47172445,  0.1418827 ,  0.82145924],
            [ 0.57585845,  0.24674534,  0.93486666],
            [ 0.53592933,  0.19119434,  0.89493754],
            [ 0.48995122,  0.16342433,  0.84895943],
            [ 0.45271317,  0.12618628,  0.80713767],
            [ 0.49519413,  0.16535238,  0.84492892],
            [ 0.27167824, -0.05743487,  0.63068646],
            [ 0.27813822, -0.06659677,  0.63714644],
            [ 0.31569424, -0.01083265,  0.67470245],
            [ 0.3983224 ,  0.07179551,  0.75274689],
            [ 0.34822681,  0.01838506,  0.69796159],
            [ 0.62098686,  0.29187374,  0.97999507],
            [ 0.47190757,  0.12717257,  0.83091578]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=33, stop=50, step=1)
    )

    forecaster = ForecasterDirect(
                     regressor = LinearRegression(), 
                     lags      = 3,
                     steps     = 8
                 )
    n_backtest = 20
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 3,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_interval_yes_exog_not_allow_remainder_gap_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Regressor is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    20 observations to backtest, steps=5 and gap=3, metric='mean_squared_error',
    'use_in_sample_residuals = True', allow_incomplete_fold = False
    """
    y_with_index = y.copy()
    y_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_with_index = exog.copy()
    exog_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')

    expected_metric = pd.DataFrame({"mean_squared_error": [0.08826806818628832]})
    expected_predictions = pd.DataFrame(
        data = np.array([[ 0.65022999,  0.32111688,  1.0092382 ],
                        [ 0.52364637,  0.17891138,  0.88265459],
                        [ 0.46809256,  0.14156566,  0.82710077],
                        [ 0.48759732,  0.16107043,  0.84202181],
                        [ 0.47172445,  0.1418827 ,  0.82145924],
                        [ 0.57585845,  0.24674534,  0.93486666],
                        [ 0.53592933,  0.19119434,  0.89493754],
                        [ 0.48995122,  0.16342433,  0.84895943],
                        [ 0.45271317,  0.12618628,  0.80713767],
                        [ 0.49519413,  0.16535238,  0.84492892],
                        [ 0.27167824, -0.05743487,  0.63068646],
                        [ 0.27813822, -0.06659677,  0.63714644],
                        [ 0.31569424, -0.01083265,  0.67470245],
                        [ 0.3983224 ,  0.07179551,  0.75274689],
                        [ 0.34822681,  0.01838506,  0.69796159]]),
        columns = ['pred', 'lower_bound', 'upper_bound'],
        index = pd.date_range(start='2022-02-03', periods=15, freq='D')
    )

    forecaster = ForecasterDirect(
                     regressor = LinearRegression(), 
                     lags      = 3,
                     steps     = 8
                 )
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_with_index) - 20,
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 3,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y_with_index,
                                       exog                    = exog_with_index,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )
    backtest_predictions = backtest_predictions.asfreq('D')

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Return predictors                                                          *
# ******************************************************************************


def test_output_backtesting_forecaster_return_predictors_same_predictions_as_predict_ForecasterRecursive():
    """
    Test predictions from _backtesting_forecaster predictors are the same as
    predictions from predict method.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.05585411566592716]})
    
    forecaster = ForecasterRecursive(regressor=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps              = 4,
            initial_train_size = len(y_train),
            refit              = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster        = forecaster,
                                        y                 = y,
                                        exog              = exog,
                                        cv                = cv,
                                        metric            = 'mean_squared_error',
                                        return_predictors = True,
                                        verbose           = False
                                   )
    
    forecaster.fit(y=y_train, exog=exog[:len(y_train)])
    expected_predictions = forecaster.regressor.predict(
        backtest_predictions[forecaster.X_train_features_names_out_]
    )

    pd.testing.assert_frame_equal(expected_metric, metric)
    np.testing.assert_array_almost_equal(
        expected_predictions, 
        backtest_predictions['pred'].to_numpy()
    )


def test_output_backtesting_forecaster_return_predictors_same_predictions_as_predict_ForecasterDirect():
    """
    Test predictions from _backtesting_forecaster predictors are the same as
    predictions from predict method.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.061964730085838]})
    
    forecaster = ForecasterDirect(regressor=LinearRegression(), steps=5, lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]

    cv = TimeSeriesFold(
             steps              = 5,
             initial_train_size = len(y_train)
         )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster        = forecaster,
                                        y                 = y,
                                        exog              = exog,
                                        cv                = cv,
                                        metric            = 'mean_squared_error',
                                        return_predictors = True,
                                        verbose           = False
                                   )
    
    forecaster.fit(y=y_train, exog=exog[:len(y_train)])
    regressors = [1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 1, 2]
    len_predictions = len(backtest_predictions)
    results = np.full(shape=len_predictions, fill_value=np.nan, dtype=float)
    for i, step in enumerate(regressors):
        results[i] = forecaster.regressors_[step].predict(
            backtest_predictions.iloc[[i]][
                ['lag_1', 'lag_2', 'lag_3', 'exog']
            ]
        )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    np.testing.assert_array_almost_equal(results, backtest_predictions['pred'].to_numpy())



# Unit test get_feature_importances ForecasterDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import RandomForestRegressor
from lightgbm import LGBMRegressor

from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect

# Fixtures
from .fixtures_forecaster_direct import y
from .fixtures_forecaster_direct import exog


def test_TypeError_is_raised_when_step_is_not_int():
    """
    Test TypeError is raised when calling get_feature_importances() and step is 
    not an int.
    """
    forecaster = ForecasterDirect(
                     regressor = RandomForestRegressor(random_state=123),
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(y=pd.Series(np.arange(5)))

    step = 'not_an_int'

    err_msg = re.escape(f"`step` must be an integer. Got {type(step)}.")
    with pytest.raises(TypeError, match = err_msg):         
        forecaster.get_feature_importances(step=step)


def test_NotFittedError_is_raised_when_forecaster_is_not_fitted():
    """
    Test NotFittedError is raised when calling get_feature_importances() and 
    forecaster is not fitted.
    """
    forecaster = ForecasterDirect(
                     regressor = RandomForestRegressor(random_state=123),
                     lags      = 3,
                     steps     = 1
                 )

    err_msg = re.escape(
        "This forecaster is not fitted yet. Call `fit` with appropriate "
        "arguments before using `get_feature_importances()`."
    )
    with pytest.raises(NotFittedError, match = err_msg):         
        forecaster.get_feature_importances(step=1)


@pytest.mark.parametrize("step", [0, 2], ids=lambda step: f'step: {step}')
def test_exception_is_raised_when_step_is_greater_than_forecaster_max_step(step):
    """
    Test exception is raised when calling get_feature_importances() and step is 
    less than 1 or greater than the forecaster.max_step.
    """
    forecaster = ForecasterDirect(
                     regressor = RandomForestRegressor(random_state=123),
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(y=pd.Series(np.arange(5)))

    err_msg = re.escape(
        f"The step must have a value from 1 to the maximum number of steps "
        f"({forecaster.max_step}). Got {step}."
    )
    with pytest.raises(ValueError, match = err_msg):         
        forecaster.get_feature_importances(step=step)


def test_output_get_feature_importances_when_regressor_is_RandomForestRegressor_lags_3_step_1():
    """
    Test output of get_feature_importances for step 1, when regressor is RandomForestRegressor with lags=3
    and it is trained with y=pd.Series(np.arange(5)).
    """
    forecaster = ForecasterDirect(
                     regressor = RandomForestRegressor(random_state=123),
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(y=pd.Series(np.arange(5)))

    results = forecaster.get_feature_importances(step=1)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3'],
                   'importance': np.array([0.3902439024390244, 0.3170731707317073, 0.2926829268292683])
               }).sort_values(by='importance', ascending=False)
    
    pd.testing.assert_frame_equal(results, expected)
  
    
def test_output_get_feature_importances_when_regressor_is_RandomForestRegressor_lags_3_step_2_exog_included():
    """
    Test output of get_feature_importances for step 2, when regressor is 
    RandomForestRegressor with lags=3, steps 3 and it is trained with 
    y pandas Series and exog is pandas DataFrame.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': np.arange(100, 110, dtype=float),
                         'exog_2': np.arange(1000, 1010, dtype=float)})
    
    forecaster = ForecasterDirect(
                     regressor = RandomForestRegressor(n_estimators=5, max_depth=2, random_state=123),
                     lags      = 3,
                     steps     = 3
                 )
    forecaster.fit(y=y, exog=exog)

    results = forecaster.get_feature_importances(step=2)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', 'exog_1', 'exog_2'],
                   'importance': np.array([0.16428571, 0.2, 0.20446429, 0.23333333, 0.19791667])
               }).sort_values(by='importance', ascending=False)
    
    pd.testing.assert_frame_equal(results, expected)
    
    
def test_output_get_feature_importances_when_regressor_is_LinearRegression_lags_3_step_1():
    """
    Test output of get_feature_importances for step 1, when regressor is LinearRegression with lags=3
    and it is trained with y=pd.Series(np.arange(5)).
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=1)
    forecaster.fit(y=pd.Series(np.arange(5)))

    results = forecaster.get_feature_importances(step=1, sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3'],
                   'importance': np.array([0.33333333, 0.33333333, 0.33333333])
               })
    
    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_LinearRegression_lags_3_step_1_exog_included():
    """
    Test output of get_feature_importances for step 1, when regressor is LinearRegression with lags=3
    and it is trained with y=pd.Series(np.arange(5)) and
    exog=pd.Series(np.arange(5), name='exog').
    """
    forecaster = ForecasterDirect(regressor=LinearRegression(), lags=3, steps=1)
    forecaster.fit(y=pd.Series(np.arange(5)), exog=pd.Series(np.arange(5), name='exog'))

    results = forecaster.get_feature_importances(step=1, sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', 'exog'],
                   'importance': np.array([0.25, 0.25, 0.25, 0.25])
               })
    
    pd.testing.assert_frame_equal(results, expected)
    

def test_output_get_feature_importances_when_regressor_no_attributes():
    """
    Test output of get_feature_importances when regressor is MLPRegressor with lags=5
    and it is trained with y=pd.Series(np.arange(10)). Since MLPRegressor hasn't attributes
    `feature_importances_` or `coef_, results = None and a warning is raised`
    """
    forecaster = ForecasterDirect(
                     regressor = MLPRegressor(solver='lbfgs', max_iter=50, random_state=123),
                     lags      = 5,
                     steps     = 1
                 )
    forecaster.fit(y=pd.Series(np.arange(10)))

    estimator = forecaster.regressor
    expected = None

    warn_msg = re.escape(
            (f"Impossible to access feature importances for regressor of type "
             f"{type(estimator)}. This method is only valid when the "
             f"regressor stores internally the feature importances in the "
             f"attribute `feature_importances_` or `coef_`.")
        )
    with pytest.warns(UserWarning, match = warn_msg):
        results = forecaster.get_feature_importances(step=1)
        assert results is expected


def test_output_get_feature_importances_when_pipeline_LinearRegression():
    """
    Test output of get_feature_importances when regressor is pipeline,
    (StandardScaler() + LinearRegression with lags=3),
    it is trained with y=pd.Series(np.arange(5)).
    """
    forecaster = ForecasterDirect(
                     regressor = make_pipeline(StandardScaler(), LinearRegression()),
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(y=pd.Series(np.arange(5)))
    
    results = forecaster.get_feature_importances(step=1, sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3'],
                   'importance': np.array([0.166667, 0.166667, 0.166667])
               })
    
    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_pipeline_RandomForestRegressor():
    """
    Test output of get_feature_importances when regressor is pipeline,
    (StandardScaler() + RandomForestRegressor with lags=3),
    it is trained with y=pd.Series(np.arange(5)).
    """
    forecaster = ForecasterDirect(
                     regressor = make_pipeline(StandardScaler(), RandomForestRegressor(n_estimators=5, max_depth=2, random_state=123)),
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(y=pd.Series(np.arange(5)))
    
    results = forecaster.get_feature_importances(step=1, sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3'],
                   'importance': np.array([0.5, 0.5, 0.0])
               })
    
    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_window_features():
    """
    Test output of get_feature_importances when regressor is LGMBRegressor with 
    lags=3 and window features.
    """
    y_datetime = y.copy()
    y_datetime.index = pd.date_range(start='2001-01-01', periods=len(y_datetime), freq='D')
    exog_datetime = exog.copy()
    exog_datetime.index = pd.date_range(start='2001-01-01', periods=len(exog_datetime), freq='D')
    
    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterDirect(
                     regressor       = LGBMRegressor(verbose=-1, random_state=123),
                     steps           = 3,
                     lags            = 3,
                     window_features = rolling
                 )
    forecaster.fit(y=y_datetime, exog=exog_datetime)

    results = forecaster.get_feature_importances(step=2, sort_importance=False)
    results = results.astype({'importance': float})
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', 'roll_mean_3', 'roll_sum_5', 'exog'],
                   'importance': np.array([13., 23., 1., 17., 33., 13.])
               }).astype({'importance': float})
    
    pd.testing.assert_frame_equal(results, expected)



################################################################################
#                                ForecasterRnn                                 #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License      #
################################################################################
# coding=utf-8

from __future__ import annotations

import sys
import warnings
from copy import deepcopy
from typing import Any

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.base import clone
from sklearn.preprocessing import MinMaxScaler

import skforecast
from ..base import ForecasterBase
from ..exceptions import DataTransformationWarning
from ..utils import (
    check_exog,
    check_interval,
    check_predict_input,
    check_residuals_input,
    check_select_fit_kwargs,
    check_y,
    check_extract_values_and_index,
    expand_index,
    get_exog_dtypes,
    get_style_repr_html,
    initialize_lags,
    initialize_transformer_series,
    input_to_frame,
    prepare_levels_multiseries,
    prepare_steps_direct,
    set_skforecast_warnings,
    transform_dataframe,
    transform_numpy,
    transform_series,
    check_optional_dependency
)

try:
    import keras
except ImportError as e:
    import sys
    if sys.version_info >= (3, 13):
        raise ImportError(
            "Python 3.13 is not supported by TensorFlow, which is the default "
            "backend used by Keras. To use Keras with Python 3.13, the KERAS_BACKEND "
            "environment variable needs to be set to 'torch', `os.environ['KERAS_BACKEND'] = 'torch'`."
            "Make sure you have PyTorch installed to use Keras with the torch backend. "
            "For installation instructions, visit https://pytorch.org/get-started/locally/"
        )
    else:
        package_name = str(e).split(" ")[-1].replace("'", "")
        check_optional_dependency(package_name=package_name)


# TODO. Include window features
# TODO. Include differentiation
# TODO. Include binner residuals
class ForecasterRnn(ForecasterBase):
    """
    This class turns any regressor compatible with the Keras API into a
    Keras RNN multi-series multi-step forecaster. A unique model is created
    to forecast all time steps and series. Keras enables workflows on top of
    either JAX, TensorFlow, or PyTorch. See documentation for more details.

    Parameters
    ----------
    regressor : regressor or pipeline compatible with the Keras API
        An instance of a regressor or pipeline compatible with the Keras API.
    levels : str, list
        Name of one or more time series to be predicted. This determine the series
        the forecaster will be handling. If `None`, all series used during training
        will be available for prediction.
    lags : int, list, numpy ndarray, range
        Lags used as predictors. Index starts at 1, so lag 1 is equal to t-1.
    
        - `int`: include lags from 1 to `lags` (included).
        - `list`, `1d numpy ndarray` or `range`: include only lags present in 
        `lags`, all elements must be int.
    transformer_series : transformer (preprocessor), dict, default None
        An instance of a transformer (preprocessor) compatible with the scikit-learn
        preprocessing API with methods: fit, transform, fit_transform and 
        inverse_transform. Transformation is applied to each `series` before training 
        the forecaster. ColumnTransformers are not allowed since they do not have 
        inverse_transform method.

        - If single transformer: it is cloned and applied to all series. 
        - If `dict` of transformers: a different transformer can be used for each series.
    transformer_exog : transformer, default None
        An instance of a transformer (preprocessor) compatible with the scikit-learn
        preprocessing API. The transformation is applied to `exog` before training the
        forecaster. `inverse_transform` is not available when using ColumnTransformers.
    fit_kwargs : dict, default `None`
        Additional arguments to be passed to the `fit` method of the regressor.
    forecaster_id : str, int, default `None`
        Name used as an identifier of the forecaster.

    Attributes
    ----------
    regressor : regressor or pipeline compatible with the Keras API
        An instance of a regressor or pipeline compatible with the Keras API.
        An instance of this regressor is trained for each step. All of them
        are stored in `self.regressors_`.
    levels : str, list
        Name of one or more time series to be predicted. This determine the series
        the forecaster will be handling. If `None`, all series used during training
        will be available for prediction.
    layers_names : list
        Names of the layers in the Keras model used as regressor.
    steps : numpy ndarray
        Future steps the forecaster will predict when using prediction methods.
    max_step : int
        Maximum step the forecaster is able to predict. It is the maximum value
        included in `steps`.
    lags : numpy ndarray
        Lags used as predictors.
    max_lag : int
        Maximum lag included in `lags`.
    window_size : int
        Size of the window needed to create the predictors.
    transformer_series : object, dict
        An instance of a transformer (preprocessor) compatible with the scikit-learn
        preprocessing API with methods: fit, transform, fit_transform and
        inverse_transform. Transformation is applied to each `series` before training
        the forecaster. ColumnTransformers are not allowed since they do not have
        inverse_transform method.
    transformer_series_ : dict
        Dictionary with the transformer for each series. It is created cloning the
        objects in `transformer_series` and is used internally to avoid overwriting.
    transformer_exog : transformer
        An instance of a transformer (preprocessor) compatible with the scikit-learn
        preprocessing API. The transformation is applied to `exog` before training the
        forecaster. `inverse_transform` is not available when using ColumnTransformers.
    last_window_ : pandas DataFrame
        This window represents the most recent data observed by the predictor
        during its training phase. It contains the values needed to predict the
        next step immediately after the training data. These values are stored
        in the original scale of the time series before undergoing any transformations
        or differentiation.
    index_type_ : type
        Type of index of the input used in training.
    index_freq_ : str
        Frequency of Index of the input used in training.
    training_range_: pandas Index
        First and last values of index of the data used during training.
    series_names_in_ : list
        Names of the series used during training.
    n_series_in : int
        Number of series used during training.
    n_levels_out : int
        Number of levels (series) to be predicted by the forecaster.
    exog_in_ : bool
        If the forecaster has been trained using exogenous variable/s.
    exog_names_in_ : list
        Names of the exogenous variables used during training.
    exog_type_in_ : type
        Type of exogenous variable/s used in training.
    exog_dtypes_in_ : dict
        Type of each exogenous variable/s used in training before the transformation
        applied by `transformer_exog`. If `transformer_exog` is not used, it
        is equal to `exog_dtypes_out_`.
    exog_dtypes_out_ : dict
        Type of each exogenous variable/s used in training after the transformation 
        applied by `transformer_exog`. If `transformer_exog` is not used, it 
        is equal to `exog_dtypes_in_`.
    X_train_dim_names_ : dict
        Labels for the multi-dimensional arrays created internally for training.
    y_train_dim_names_ : dict
        Labels for the multi-dimensional arrays created internally for training.
    series_val : pandas DataFrame
        Values of the series used for validation during training.
    exog_val : pandas DataFrame
        Values of the exogenous variables used for validation during training.
    history : dict
        Dictionary with the history of the training of each step. It is created
        internally to avoid overwriting.
    fit_kwargs : dict
        Additional arguments to be passed to the `fit` method of the regressor.
    in_sample_residuals_ : dict
        Residuals of the model when predicting training data. Only stored up 
        to 10_000 values per step in the form `{step: residuals}`. If 
        `transformer_series` is not `None`, residuals are stored in the 
        transformed scale.
    out_sample_residuals_ : dict
        Residuals of the model when predicting non-training data. Only stored up 
        to 10_000 values per step in the form `{step: residuals}`. Use 
        `set_out_sample_residuals()` method to set values. If `transformer_series` 
        is not `None`, residuals are stored in the transformed scale.
    creation_date : str
        Date of creation.
    is_fitted : bool
        Tag to identify if the regressor has been fitted (trained).
    fit_date : str
        Date of last fit.
    keras_backend_ : str
        Keras backend used to fit the forecaster. It can be 'tensorflow', 'torch' 
        or 'jax'.
    skforecast_version : str
        Version of skforecast library used to create the forecaster.
    python_version : str
        Version of python used to create the forecaster.
    forecaster_id : str, int
        Name used as an identifier of the forecaster.
    _probabilistic_mode: str, bool
        Private attribute used to indicate whether the forecaster should perform 
        some calculations during backtesting.
    dropna_from_series : Ignored
        Not used, present here for API consistency by convention.
    encoding : Ignored
        Not used, present here for API consistency by convention.
    differentiation : Ignored
        Not used, present here for API consistency by convention.
    differentiation_max : Ignored
        Not used, present here for API consistency by convention.
    differentiator : Ignored
        Not used, present here for API consistency by convention.
    differentiator_ : Ignored
        Not used, present here for API consistency by convention.

    """

    def __init__(
        self,
        regressor: object,
        levels: str | list[str],
        lags: int | list[int] | np.ndarray[int] | range[int],
        transformer_series: object | dict[str, object] | None = MinMaxScaler(
            feature_range=(0, 1)
        ),
        transformer_exog: object | None = MinMaxScaler(feature_range=(0, 1)),
        fit_kwargs: dict[str, object] | None = {},
        forecaster_id: str | int | None = None
    ) -> None:
        
        self.regressor = deepcopy(regressor)
        self.levels = None
        self.transformer_series = transformer_series
        self.transformer_series_ = None
        self.transformer_exog = transformer_exog
        self.max_lag = None
        self.window_size = None
        self.last_window_ = None
        self.index_type_ = None
        self.index_freq_ = None
        self.training_range_ = None
        self.series_names_in_ = None
        self.exog_names_in_ = None
        self.exog_type_in_ = None
        self.exog_dtypes_in_ = None
        self.X_train_dim_names_ = None
        self.y_train_dim_names_ = None
        self.history_ = None 
        self.is_fitted = False
        self.creation_date = pd.Timestamp.today().strftime("%Y-%m-%d %H:%M:%S")
        self.fit_date = None
        self.keras_backend_ = None
        self.skforecast_version = skforecast.__version__
        self.python_version = sys.version.split(" ")[0]
        self.forecaster_id = forecaster_id
        self._probabilistic_mode = "no_binned"
        self.weight_func = None  # Ignored in this forecaster
        self.source_code_weight_func = None  # Ignored in this forecaster
        self.dropna_from_series = False  # Ignored in this forecaster
        self.encoding = None  # Ignored in this forecaster
        self.differentiation = None  # Ignored in this forecaster
        self.differentiation_max = None  # Ignored in this forecaster
        self.differentiator = None  # Ignored in this forecaster
        self.differentiator_ = None  # Ignored in this forecaster

        layer_init = self.regressor.layers[0]
        layer_end = self.regressor.layers[-1]
        self.layers_names = [layer.name for layer in self.regressor.layers]
        
        self.lags, self.lags_names, self.max_lag = initialize_lags(
            type(self).__name__, lags
        )
        n_lags_regressor = layer_init.output.shape[1]
        if len(self.lags) != n_lags_regressor:
            raise ValueError(
                f"Number of lags ({len(self.lags)}) does not match the number of "
                f"lags expected by the regressor architecture ({n_lags_regressor})."
            )
        
        self.window_size = self.max_lag

        self.steps = np.arange(layer_end.output.shape[1]) + 1
        self.max_step = np.max(self.steps)

        if isinstance(levels, str):
            self.levels = [levels]
        elif isinstance(levels, list):
            self.levels = levels
        else:
            raise TypeError(
                f"`levels` argument must be a string or list. Got {type(levels)}."
            )
        
        self.n_series_in = self.regressor.get_layer('series_input').output.shape[-1]
        self.n_levels_out = self.regressor.get_layer('output_dense_td_layer').output.shape[-1]
        self.exog_in_ = True if "exog_input" in self.layers_names else False
        if self.exog_in_:
            self.n_exog_in = self.regressor.get_layer('exog_input').output.shape[-1]
        else:
            self.n_exog_in = None
            # NOTE: This is needed because the Reshape layer changes the output 
            # shape in _create_and_compile_model_no_exog
            self.n_levels_out = int(self.n_levels_out / self.max_step)

        if not len(self.levels) == self.n_levels_out:
            raise ValueError(
                f"Number of levels ({len(self.levels)}) does not match the number of "
                f"levels expected by the regressor architecture ({self.n_levels_out})."
            )

        self.series_val = None
        self.exog_val = None
        if "series_val" in fit_kwargs:
            if not isinstance(fit_kwargs["series_val"], pd.DataFrame):
                raise TypeError(
                    f"`series_val` must be a pandas DataFrame. "
                    f"Got {type(fit_kwargs['series_val'])}."
                )
            self.series_val = fit_kwargs.pop("series_val")            

            if self.exog_in_:
                if "exog_val" not in fit_kwargs.keys():
                    raise ValueError(
                        "If `series_val` is provided, `exog_val` must also be "
                        "provided using the `fit_kwargs` argument when the "
                        "regressor has exogenous variables."
                    )
                else:
                    if not isinstance(fit_kwargs["exog_val"], (pd.Series, pd.DataFrame)):
                        raise TypeError(
                            f"`exog_val` must be a pandas Series or DataFrame. "
                            f"Got {type(fit_kwargs['exog_val'])}."
                        )
                    self.exog_val = input_to_frame(
                        data=fit_kwargs.pop("exog_val"), input_name='exog_val'
                    )

        self.in_sample_residuals_ = None
        self.in_sample_residuals_by_bin_ = None  # Ignored in this forecaster
        self.out_sample_residuals_ = None
        self.out_sample_residuals_by_bin_ = None  # Ignored in this forecaster

        self.fit_kwargs = check_select_fit_kwargs(
            regressor=self.regressor, fit_kwargs=fit_kwargs
        )

    def __repr__(self) -> str:
        """
        Information displayed when a ForecasterRnn object is printed.
        """

        params = str(self.regressor.get_config())
        compile_config = str(self.regressor.get_compile_config())
        
        (
            _,
            _,
            series_names_in_,
            exog_names_in_,
            transformer_series,
        ) = [
            self._format_text_repr(value) 
            for value in self._preprocess_repr(
                regressor          = None,
                series_names_in_   = self.series_names_in_,
                exog_names_in_     = self.exog_names_in_,
                transformer_series = self.transformer_series,
            )
        ]

        info = (
            f"{'=' * len(type(self).__name__)} \n"
            f"{type(self).__name__} \n"
            f"{'=' * len(type(self).__name__)} \n"
            f"Regressor: {self.regressor} \n"
            f"Layers names: {self.layers_names} \n"
            f"Lags: {self.lags} \n"
            f"Window size: {self.window_size} \n"
            f"Maximum steps to predict: {self.steps} \n"
            f"Series names: {series_names_in_} \n"
            f"Target series (levels): {self.levels} \n"
            f"Exogenous included: {self.exog_in_} \n"
            f"Exogenous names: {exog_names_in_} \n"
            f"Transformer for series: {transformer_series} \n"
            f"Transformer for exog: {self.transformer_exog} \n"
            f"Training range: {self.training_range_.to_list() if self.is_fitted else None} \n"
            f"Training index type: {str(self.index_type_).split('.')[-1][:-2] if self.is_fitted else None} \n"
            f"Training index frequency: {self.index_freq_ if self.is_fitted else None} \n"
            f"Regressor parameters: {params} \n"
            f"Compile parameters: {compile_config} \n"
            f"fit_kwargs: {self.fit_kwargs} \n"
            f"Creation date: {self.creation_date} \n"
            f"Last fit date: {self.fit_date} \n"
            f"Keras backend: {self.keras_backend_} \n"
            f"Skforecast version: {self.skforecast_version} \n"
            f"Python version: {self.python_version} \n"
            f"Forecaster id: {self.forecaster_id} \n"
        )

        return info

    def _repr_html_(self):
        """
        HTML representation of the object.
        The "General Information" section is expanded by default.
        """

        params = str(self.regressor.get_config())
        compile_config = str(self.regressor.get_compile_config())

        (
            _,
            _,
            series_names_in_,
            exog_names_in_,
            transformer_series,
        ) = self._preprocess_repr(
                regressor          = None,
                series_names_in_   = self.series_names_in_,
                exog_names_in_     = self.exog_names_in_,
                transformer_series = self.transformer_series,
            )

        style, unique_id = get_style_repr_html(self.is_fitted)
        
        content = f"""
        <div class="container-{unique_id}">
            <h2>{type(self).__name__}</h2>
            <details open>
                <summary>General Information</summary>
                <ul>
                    <li><strong>Regressor:</strong> {type(self.regressor).__name__}</li>
                    <li><strong>Layers names:</strong> {self.layers_names}</li>
                    <li><strong>Lags:</strong> {self.lags}</li>
                    <li><strong>Window size:</strong> {self.window_size}</li>
                    <li><strong>Maximum steps to predict:</strong> {self.steps}</li>
                    <li><strong>Exogenous included:</strong> {self.exog_in_}</li>
                    <li><strong>Creation date:</strong> {self.creation_date}</li>
                    <li><strong>Last fit date:</strong> {self.fit_date}</li>
                    <li><strong>Keras backend:</strong> {self.keras_backend_}</li>
                    <li><strong>Skforecast version:</strong> {self.skforecast_version}</li>
                    <li><strong>Python version:</strong> {self.python_version}</li>
                    <li><strong>Forecaster id:</strong> {self.forecaster_id}</li>
                </ul>
            </details>
            <details>
                <summary>Exogenous Variables</summary>
                <ul>
                    {exog_names_in_}
                </ul>
            </details>
            <details>
                <summary>Data Transformations</summary>
                <ul>
                    <li><strong>Transformer for series:</strong> {transformer_series}</li>
                    <li><strong>Transformer for exog:</strong> {self.transformer_exog}</li>
                </ul>
            </details>
            <details>
                <summary>Training Information</summary>
                <ul>
                    <li><strong>Series names:</strong> {series_names_in_}</li>
                    <li><strong>Target series (levels):</strong> {self.levels}</li>
                    <li><strong>Training range:</strong> {self.training_range_.to_list() if self.is_fitted else 'Not fitted'}</li>
                    <li><strong>Training index type:</strong> {str(self.index_type_).split('.')[-1][:-2] if self.is_fitted else 'Not fitted'}</li>
                    <li><strong>Training index frequency:</strong> {self.index_freq_ if self.is_fitted else 'Not fitted'}</li>
                </ul>
            </details>
            <details>
                <summary>Regressor Parameters</summary>
                <ul>
                    {params}
                </ul>
            </details>
            <details>
                <summary>Compile Parameters</summary>
                <ul>
                    {compile_config}
                </ul>
            </details>
            <details>
                <summary>Fit Kwargs</summary>
                <ul>
                    {self.fit_kwargs}
                </ul>
            </details>
            <p>
                <a href="https://skforecast.org/{skforecast.__version__}/api/forecasterrnn.html">&#128712 <strong>API Reference</strong></a>
                &nbsp;&nbsp;
                <a href="https://skforecast.org/{skforecast.__version__}/user_guides/forecasting-with-deep-learning-rnn-lstm.html">&#128462 <strong>User Guide</strong></a>
            </p>
        </div>
        """

        # Return the combined style and content
        return style + content

    # TODO: CREATE_LAGS_AND_STEPS
    def _create_lags(
        self, 
        y: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Transforms a 1d array into a 3d array (X) and a 3d array (y). Each row
        in X is associated with a value of y and it represents the lags that
        precede it.

        Notice that, the returned matrix X_data, contains the lag 1 in the first
        column, the lag 2 in the second column and so on.

        Parameters
        ----------
        y : numpy ndarray
            1d numpy ndarray Training time series.

        Returns
        -------
        X_data : numpy ndarray
            3d numpy ndarray with the lagged values (predictors).
            Shape: (samples - max(lags), len(lags))
        y_data : numpy ndarray
            3d numpy ndarray with the values of the time series related to each
            row of `X_data` for each step.
            Shape: (len(max_step), samples - max(lags))

        """

        n_rows = len(y) - self.window_size - self.max_step + 1  # rows of y_data

        X_data = np.full(
            shape=(n_rows, (self.window_size)), fill_value=np.nan, order="F", dtype=float
        )
        for i, lag in enumerate(range(self.window_size - 1, -1, -1)):
            X_data[:, i] = y[self.window_size - lag - 1 : -(lag + self.max_step)]

        # Get lags index
        X_data = X_data[:, self.lags - 1]

        y_data = np.full(
            shape=(n_rows, self.max_step), fill_value=np.nan, order="F", dtype=float
        )
        for step in range(self.max_step):
            y_data[:, step] = y[self.window_size + step : self.window_size + step + n_rows]

        # Get steps index
        y_data = y_data[:, self.steps - 1]

        return X_data, y_data

    def _create_train_X_y(
        self,
        series: pd.DataFrame,
        exog: pd.Series | pd.DataFrame | None = None
    ) -> tuple[
        np.ndarray, 
        np.ndarray, 
        np.ndarray, 
        dict[int, list], 
        list[str], 
        dict[str, type], 
        dict[str, type]
    ]:
        """
        Create training matrices. The resulting multi-dimensional matrices contain
        the target variable and predictors needed to train the model.

        Parameters
        ----------
        series : pandas DataFrame
            Training time series.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s. Must have the same
            number of observations as `series` and their indexes must be aligned.

        Returns
        -------
        X_train : numpy ndarray
            Training values (predictors) for each step. The resulting array has
            3 dimensions: (n_observations, n_lags, n_series)
        exog_train: numpy ndarray
            Value of exogenous variables aligned with X_train. (n_observations, n_exog)
        y_train : numpy ndarray
            Values (target) of the time series related to each row of `X_train`.
            The resulting array has 3 dimensions: (n_observations, n_steps, n_levels)
        dimension_names : dict
            Labels for the multi-dimensional arrays created internally for training.
        exog_names_in_ : list
            Names of the exogenous variables included in the training matrices.
        exog_dtypes_in_ : dict
            Type of each exogenous variable/s used in training before the transformation
            applied by `transformer_exog`. If `transformer_exog` is not used, it
            is equal to `exog_dtypes_out_`.
        exog_dtypes_out_ : dict
            Type of each exogenous variable/s used in training after the transformation 
            applied by `transformer_exog`. If `transformer_exog` is not used, it 
            is equal to `exog_dtypes_in_`.

        """

        if not isinstance(series, pd.DataFrame):
            raise TypeError(f"`series` must be a pandas DataFrame. Got {type(series)}.")

        _, series_index = check_extract_values_and_index(
            data=series, data_label="`series`", return_values=False
        )
        series_names_in_ = list(series.columns)
        if not len(series_names_in_) == self.n_series_in:
            raise ValueError(
                f"Number of series in `series` ({len(series_names_in_)}) "
                f"does not match the number of series expected by the model "
                f"architecture ({self.n_series_in})."
            )

        if not set(self.levels).issubset(set(series_names_in_)):
            raise ValueError(
                f"`levels` defined when initializing the forecaster must be "
                f"included in `series` used for training. "
                f"{set(self.levels) - set(series_names_in_)} not found."
            )

        if len(series) < self.window_size + self.max_step:
            raise ValueError(
                f"Minimum length of `series` for training this forecaster is "
                f"{self.window_size + self.max_step}. Reduce the number of "
                f"predicted steps, {self.max_step}, or the maximum "
                f"lag, {self.max_lag}, if no more data is available.\n"
                f"    Length `series`: {len(series)}.\n"
                f"    Max step : {self.max_step}.\n"
                f"    Lags window size: {self.max_lag}."
            )
        
        if exog is None and self.exog_in_:
            raise ValueError(
                "The regressor architecture expects exogenous variables during "
                "training. Please provide the `exog` argument. If this is "
                "unexpected, check your regressor architecture or the "
                "initialization parameters of the forecaster."
            )
        if exog is not None and not self.exog_in_:
            raise ValueError(
                "Exogenous variables (`exog`) were provided, but the model "
                "architecture was not built to expect exogenous variables. Please "
                "remove the `exog` argument or rebuild the model to include "
                "exogenous inputs."
            )

        fit_transformer = False
        if not self.is_fitted:
            fit_transformer = True
            self.transformer_series_ = initialize_transformer_series(
                                           forecaster_name    = type(self).__name__,
                                           series_names_in_   = series_names_in_,
                                           transformer_series = self.transformer_series
                                       )

        # Step 1: Create lags for all columns
        X_train = []
        y_train = []

        # TODO: Add method argument to calculate lags and/or steps
        for serie in series_names_in_:
            x = series[serie]
            check_y(y=x)
            x = transform_series(
                series=x,
                transformer=self.transformer_series_[serie],
                fit=fit_transformer,
                inverse_transform=False,
            )
            X, _ = self._create_lags(x)
            X_train.append(X)

        for level in self.levels:
            y = series[level]
            check_y(y=y)
            y = transform_series(
                series=y,
                transformer=self.transformer_series_[level],
                fit=fit_transformer,
                inverse_transform=False,
            )

            _, y = self._create_lags(y)
            y_train.append(y)

        X_train = np.stack(X_train, axis=2)
        y_train = np.stack(y_train, axis=2)

        train_index = series_index[
            self.max_lag : (len(series_index) - self.max_step + 1)
        ]
        dimension_names = {
            "X_train": {
                0: train_index,
                1: self.lags_names[::-1],
                2: series_names_in_,
            },
            "y_train": {
                0: train_index,
                1: [f"step_{step}" for step in self.steps],
                2: self.levels,
            },
        }

        if exog is not None:

            check_exog(exog=exog, allow_nan=False)
            exog = input_to_frame(data=exog, input_name='exog')
            _, exog_index = check_extract_values_and_index(
                data=exog, data_label='`exog`', ignore_freq=True, return_values=False
            )

            if len(exog.columns) != self.n_exog_in:
                raise ValueError(
                    f"Number of columns in `exog` ({len(exog.columns)}) "
                    f"does not match the number of exogenous variables expected "
                    f"by the model architecture ({self.n_exog_in})."
                )
            
            series_index_no_ws = series_index[self.window_size:]
            len_series = len(series)
            len_series_no_ws = len_series - self.window_size
            len_exog = len(exog)
            if not len_exog == len_series and not len_exog == len_series_no_ws:
                raise ValueError(
                    f"Length of `exog` must be equal to the length of `series` (if "
                    f"index is fully aligned) or length of `series` - `window_size` "
                    f"(if `exog` starts after the first `window_size` values).\n"
                    f"    `exog`                   : ({exog_index[0]} -- {exog_index[-1]})  (n={len_exog})\n"
                    f"    `series`                 : ({series.index[0]} -- {series.index[-1]})  (n={len_series})\n"
                    f"    `series` - `window_size` : ({series_index_no_ws[0]} -- {series_index_no_ws[-1]})  (n={len_series_no_ws})"
                )
            
            exog_names_in_ = exog.columns.to_list()
            if len(set(exog_names_in_) - set(series_names_in_)) != len(exog_names_in_):
                raise ValueError(
                    f"`exog` cannot contain a column named the same as one of "
                    f"the series (column names of series).\n"
                    f"  `series` columns : {series_names_in_}.\n"
                    f"  `exog`   columns : {exog_names_in_}."
                )
            
            exog_n_dim_in = len(exog_names_in_)
            exog_dtypes_in_ = get_exog_dtypes(exog=exog)
            exog = transform_dataframe(
                df=exog,
                transformer=self.transformer_exog,
                fit=fit_transformer,
                inverse_transform=False,
            )
            exog_n_dim_out = len(exog.columns)
            exog_dtypes_out_ = get_exog_dtypes(exog=exog)

            if exog_n_dim_in != exog_n_dim_out:
                raise ValueError(
                    f"Number of columns in `exog` after transformation ({exog_n_dim_out}) "
                    f"does not match the number of columns before transformation ({exog_n_dim_in}). "
                    f"The ForecasterRnn does not support transformations that "
                    f"change the number of columns in `exog`. Preprocess `exog` "
                    f"before passing it to the `create_and_compile_model` function."
                )

            if len_exog == len_series:
                if not (exog_index == series_index).all():
                    raise ValueError(
                        "When `exog` has the same length as `series`, the index "
                        "of `exog` must be aligned with the index of `series` "
                        "to ensure the correct alignment of values."
                    )
            else:
                if not (exog_index == series_index_no_ws).all():
                    raise ValueError(
                        "When `exog` doesn't contain the first `window_size` "
                        "observations, the index of `exog` must be aligned with "
                        "the index of `series` minus the first `window_size` "
                        "observations to ensure the correct alignment of values."
                    )
            
            exog_train = []
            for _, exog_name in enumerate(exog.columns):
                _, exog_step = self._create_lags(exog[exog_name])
                exog_train.append(exog_step)
                
            exog_train = np.stack(exog_train, axis=2)

            dimension_names["exog_train"] = {
                0: train_index,
                1: [f"step_{step}" for step in self.steps],
                2: exog.columns.to_list(),
            }
        else:
            exog_train = None
            exog_names_in_ = None
            exog_dtypes_in_ = None
            exog_dtypes_out_ = None
            dimension_names["exog_train"] = {
                0: None,
                1: None,
                2: None
            }

        return (
            X_train, 
            exog_train, 
            y_train, 
            dimension_names,
            exog_names_in_,
            exog_dtypes_in_,
            exog_dtypes_out_
        )
    
    def create_train_X_y(
        self,
        series: pd.DataFrame,
        exog: pd.Series | pd.DataFrame | None = None,
        suppress_warnings: bool = False
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, dict[int, list]]:
        """
        Create training matrices. The resulting multi-dimensional matrices contain
        the target variable and predictors needed to train the model.

        Parameters
        ----------
        series : pandas DataFrame
            Training time series.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s. Must have the same
            number of observations as `series` and their indexes must be aligned.
        suppress_warnings : bool, default False
            If `True`, skforecast warnings will be suppressed during the creation
            of the training matrices. See skforecast.exceptions.warn_skforecast_categories 
            for more information.

        Returns
        -------
        X_train : numpy ndarray
            Training values (predictors) for each step. The resulting array has
            3 dimensions: (n_observations, n_lags, n_series)
        exog_train: numpy ndarray
            Value of exogenous variables aligned with X_train. (n_observations, n_exog)
        y_train : numpy ndarray
            Values (target) of the time series related to each row of `X_train`.
            The resulting array has 3 dimensions: (n_observations, n_steps, n_levels)
        dimension_names : dict
            Labels for the multi-dimensional arrays created internally for training.

        """

        set_skforecast_warnings(suppress_warnings, action='ignore')

        output = self._create_train_X_y(series=series, exog=exog)

        X_train = output[0]
        exog_train = output[1]
        y_train = output[2]
        dimension_names = output[3]
        
        set_skforecast_warnings(suppress_warnings, action='default')

        return X_train, exog_train, y_train, dimension_names

    def fit(
        self,
        series: pd.DataFrame,
        exog: pd.Series | pd.DataFrame = None,
        store_last_window: bool = True,
        store_in_sample_residuals: bool = False,
        random_state: int = 123,
        suppress_warnings: bool = False
    ) -> None:
        """
        Training Forecaster.

        Additional arguments to be passed to the `fit` method of the regressor
        can be added with the `fit_kwargs` argument when initializing the forecaster.

        Parameters
        ----------
        series : pandas DataFrame
            Training time series.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s. Must have the same
            number of observations as `series` and their indexes must be aligned so
            that series[i] is regressed on exog[i].
        store_last_window : bool, default True
            Whether or not to store the last window (`last_window_`) of training data.
        store_in_sample_residuals : bool, default False
            If `True`, in-sample residuals will be stored in the forecaster object
            after fitting (`in_sample_residuals_` attribute).
        random_state : int, default 123
            Set a seed for the random generator so that the stored sample 
            residuals are always deterministic.
        suppress_warnings : bool, default False
            If `True`, skforecast warnings will be suppressed during the prediction
            process. See skforecast.exceptions.warn_skforecast_categories for more
            information.
        
        Returns
        -------
        None

        """

        set_skforecast_warnings(suppress_warnings, action="ignore")

        # Reset values in case the forecaster has already been fitted.
        self.last_window_ = None
        self.index_type_ = None
        self.index_freq_ = None
        self.training_range_ = None
        self.series_names_in_ = None
        self.exog_names_in_ = None
        self.exog_type_in_ = None
        self.exog_dtypes_in_ = None
        self.exog_dtypes_out_ = None
        self.X_train_dim_names_ = None
        self.y_train_dim_names_ = None
        self.exog_train_dim_names_ = None
        self.in_sample_residuals_ = None
        self.is_fitted = False
        self.fit_date = None
        self.keras_backend_ = keras.backend.backend()
            
        (
            X_train,
            exog_train,
            y_train,
            dimension_names,
            exog_names_in_,
            exog_dtypes_in_,
            exog_dtypes_out_,
        ) = self._create_train_X_y(series=series, exog=exog)

        # NOTE: Need here to avoid refitting the transformer_series_ with the 
        # validation data.
        self.is_fitted = True
        series_names_in_ = dimension_names["X_train"][2]

        if self.keras_backend_ == "torch":
            
            import torch
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            print(f"Using '{self.keras_backend_}' backend with device: {device}")

            torch_device = torch.device(device)
            X_train = torch.tensor(X_train).to(torch_device)
            y_train = torch.tensor(y_train).to(torch_device)
            if exog_train is not None:
                exog_train = torch.tensor(exog_train).to(torch_device)

        if self.series_val is not None:
            series_val = self.series_val[series_names_in_]
            if exog is not None:
                exog_val = self.exog_val[exog_names_in_]
            else:
                exog_val = None
            
            X_val, exog_val, y_val, *_ = self._create_train_X_y(
                series=series_val, exog=exog_val
            )
            if self.keras_backend_ == "torch":
                X_val = torch.tensor(X_val).to(torch_device)
                y_val = torch.tensor(y_val).to(torch_device)
                if exog_val is not None:
                    exog_val = torch.tensor(exog_val).to(torch_device)

            if self.exog_val is not None:
                history = self.regressor.fit(
                    x=[X_train, exog_train],
                    y=y_train,
                    validation_data=([X_val, exog_val], y_val),
                    **self.fit_kwargs,
                )
            else:
                history = self.regressor.fit(
                    x=X_train,
                    y=y_train,
                    validation_data=(X_val, y_val),
                    **self.fit_kwargs,
                )

        else:
            history = self.regressor.fit(
                x=X_train if exog_train is None else [X_train, exog_train],
                y=y_train,
                **self.fit_kwargs,
            )

        # TODO: Include binning in the forecaster
        self.in_sample_residuals_ = {}
        if store_in_sample_residuals:

            # NOTE: Convert to numpy array if using torch backend
            if self.keras_backend_ == "torch":
                y_train = y_train.detach().cpu().numpy()

            residuals = y_train - self.regressor.predict(
                x=X_train if exog_train is None else [X_train, exog_train], verbose=0
            )

            residuals = np.concatenate(
                [residuals[:, i, :] for i, step in enumerate(self.steps)]
            )

            rng = np.random.default_rng(seed=random_state)
            for i, level in enumerate(self.levels):
                residuals_level = residuals[:, i]
                if len(residuals_level) > 10_000:
                    residuals_level = residuals_level[
                        rng.integers(low=0, high=len(residuals_level), size=10_000)
                    ]
                self.in_sample_residuals_[level] = residuals_level
        else:
            for level in self.levels:
                self.in_sample_residuals_[level] = None
        
        self.series_names_in_ = series_names_in_
        self.X_train_series_names_in_ = series_names_in_
        self.X_train_dim_names_ = dimension_names["X_train"]
        self.y_train_dim_names_ = dimension_names["y_train"]
        self.history_ = history.history

        self.fit_date = pd.Timestamp.today().strftime("%Y-%m-%d %H:%M:%S")
        self.training_range_ = series.index[[0, -1]]
        self.index_type_ = type(series.index)
        if isinstance(series.index, pd.DatetimeIndex):
            self.index_freq_ = series.index.freqstr
        else:
            self.index_freq_ = series.index.step

        if exog is not None:
            # NOTE: self.exog_in_ is determined by the regressor architecture and
            # set during initialization.
            self.exog_names_in_ = exog_names_in_
            self.exog_type_in_ = type(exog)
            self.exog_dtypes_in_ = exog_dtypes_in_
            self.exog_dtypes_out_ = exog_dtypes_out_
            self.exog_train_dim_names_ = dimension_names["exog_train"]
            self.X_train_exog_names_out_ = dimension_names["exog_train"][2]
            self.X_train_features_names_out_ = dimension_names["X_train"][1] + dimension_names["exog_train"][2]
        else:
            self.X_train_features_names_out_ = dimension_names["X_train"][1]

        if store_last_window:
            self.last_window_ = series.iloc[-self.max_lag :, :].copy()

        set_skforecast_warnings(suppress_warnings, action="default")

    def _create_predict_inputs(
        self,
        steps: int | list[int] | None = None,
        levels: str | list[str] | None = None,
        last_window: pd.Series | pd.DataFrame | None = None,
        exog: pd.Series | pd.DataFrame | None = None,
        predict_probabilistic: bool = False,
        use_in_sample_residuals: bool = True,
        check_inputs: bool = True
    ) -> tuple[list[np.ndarray], dict[str, dict], list[int], list[str], pd.Index]:
        """
        Create the inputs needed for the prediction process.
        
        Parameters
        ----------
        steps : int, list, default None
            Predict n steps. The value of `steps` must be less than or equal to the 
            value of steps defined in the regressor architecture.
        
            - If `int`: Only steps within the range of 1 to int are predicted.
            - If `list`: List of ints. Only the steps contained in the list 
            are predicted.
            - If `None`: As many steps are predicted as defined in the regressor
            architecture.
        levels : str, list, default None
            Name(s) of the time series to be predicted. It must be included
            in `levels`, defined when initializing the forecaster. If `None`, all
            all series used during training will be available for prediction.
        last_window : pandas Series, pandas DataFrame, default None
            Series values used to create the predictors (lags) needed to 
            predict `steps`.
            If `last_window = None`, the values stored in `self.last_window_` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s.
        predict_probabilistic : bool, default False
            If `True`, the necessary checks for probabilistic predictions will be 
            performed.
        use_in_sample_residuals : bool, default True
            If `True`, residuals from the training data are used as proxy of
            prediction error to create predictions. 
            If `False`, out of sample residuals (calibration) are used. 
            Out-of-sample residuals must be precomputed using Forecaster's
            `set_out_sample_residuals()` method.
        check_inputs : bool, default True
            If `True`, the input is checked for possible warnings and errors 
            with the `check_predict_input` function. This argument is created 
            for internal use and is not recommended to be changed.

        Returns
        -------
        X : list
            List of numpy arrays needed for prediction. The first element
            is the matrix of lags and the second element is the matrix of
            exogenous variables.
        X_predict_dimension_names : dict
            Labels for the multi-dimensional arrays created internally for prediction.
        steps : list
            Steps to predict.
        levels : list
            Levels (series) to predict.
        prediction_index : pandas Index
            Index of the predictions.
        
        """

        levels, _ = prepare_levels_multiseries(
            X_train_series_names_in_=self.levels, levels=levels
        )

        steps = prepare_steps_direct(
                    max_step = self.steps,
                    steps    = steps
                )

        if last_window is None:
            last_window = self.last_window_

        if check_inputs:
            check_predict_input(
                forecaster_name=type(self).__name__,
                steps=steps,
                is_fitted=self.is_fitted,
                exog_in_=self.exog_in_,
                index_type_=self.index_type_,
                index_freq_=self.index_freq_,
                window_size=self.window_size,
                last_window=last_window,
                exog=exog,
                exog_names_in_=self.exog_names_in_,
                interval=None,
                max_step=self.max_step,
                levels=levels,
                levels_forecaster=self.levels,
                series_names_in_=self.series_names_in_,
            )

            if predict_probabilistic:
                check_residuals_input(
                    forecaster_name              = type(self).__name__,
                    use_in_sample_residuals      = use_in_sample_residuals,
                    in_sample_residuals_         = self.in_sample_residuals_,
                    out_sample_residuals_        = self.out_sample_residuals_,
                    use_binned_residuals         = False,
                    in_sample_residuals_by_bin_  = None,
                    out_sample_residuals_by_bin_ = None,
                    levels                       = self.levels
                )

        last_window = last_window.iloc[
            -self.window_size :, last_window.columns.get_indexer(self.series_names_in_)
        ].copy()

        last_window_values = last_window.to_numpy()
        last_window_matrix = np.full(
            shape=last_window.shape, fill_value=np.nan, order='F', dtype=float
        )
        for idx_series, series in enumerate(self.series_names_in_):
            last_window_series = last_window_values[:, idx_series]
            last_window_series = transform_numpy(
                array=last_window_series,
                transformer=self.transformer_series_[series],
                fit=False,
                inverse_transform=False,
            )
            last_window_matrix[:, idx_series] = last_window_series

        X = [np.reshape(last_window_matrix, (1, self.max_lag, last_window.shape[1]))]
        X_predict_dimension_names = {
            "X_autoreg": {
                0: "batch",
                1: self.lags_names[::-1],
                2: self.X_train_series_names_in_
            }
        }

        if exog is not None:

            exog = input_to_frame(data=exog, input_name='exog')
            exog = transform_dataframe(
                df=exog,
                transformer=self.transformer_exog,
                fit=False,
                inverse_transform=False,
            )

            exog_pred = exog.to_numpy()[:self.max_step]

            # NOTE: This is done to ensure that the exogenous variables
            # have the same number of rows as the maximum step to predict 
            # during backtesting when the last fold is incomplete 
            if len(exog_pred) < self.max_step:
                exog_pred = np.concatenate(
                    [
                        exog_pred,
                        np.full(
                            shape=(self.max_step - len(exog_pred), exog_pred.shape[1]),
                            fill_value=0.,
                            dtype=float
                        )
                    ],
                    axis=0
                )

            exog_pred = np.expand_dims(exog_pred, axis=0)
            X.append(exog_pred)

            X_predict_dimension_names["exog_pred"] = {
                0: "batch",
                1: [f"step_{step}" for step in self.steps],
                2: self.X_train_exog_names_out_
            }
        
        prediction_index = expand_index(
                               index = last_window.index,
                               steps = max(steps)
                           )[np.array(steps) - 1]
        if isinstance(last_window.index, pd.DatetimeIndex) and np.array_equal(
            steps, np.arange(min(steps), max(steps) + 1)
        ):
            prediction_index.freq = last_window.index.freq

        return X, X_predict_dimension_names, steps, levels, prediction_index

    def create_predict_X(
        self,
        steps: int | list[int] | None = None,
        levels: str | list[str] | None = None,
        last_window: pd.DataFrame | None = None,
        exog: pd.Series | pd.DataFrame | None = None,
        suppress_warnings: bool = False,
        check_inputs: bool = True
    ) -> tuple[pd.DataFrame, pd.DataFrame | None]:
        """
        Create the predictors needed to predict `steps` ahead.
        
        Parameters
        ----------
        steps : int, list, default None
            Predict n steps. The value of `steps` must be less than or equal to the 
            value of steps defined in the regressor architecture.
        
            - If `int`: Only steps within the range of 1 to int are predicted.
            - If `list`: List of ints. Only the steps contained in the list 
            are predicted.
            - If `None`: As many steps are predicted as defined in the regressor
            architecture.
        levels : str, list, default None
            Name(s) of the time series to be predicted. It must be included
            in `levels`, defined when initializing the forecaster. If `None`, all
            all series used during training will be available for prediction.
        last_window : pandas DataFrame, default None
            Series values used to create the predictors (lags) needed to 
            predict `steps`.
            If `last_window = None`, the values stored in `self.last_window_` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s.
        suppress_warnings : bool, default False
            If `True`, skforecast warnings will be suppressed during the prediction 
            process. See skforecast.exceptions.warn_skforecast_categories for more
            information.
        check_inputs : bool, default True
            If `True`, the input is checked for possible warnings and errors 
            with the `check_predict_input` function. This argument is created 
            for internal use and is not recommended to be changed.

        Returns
        -------
        X_predict : pandas DataFrame
            Pandas DataFrame with the predictors for each step.
        exog_predict : pandas DataFrame
            Pandas DataFrame with the exogenous variables for each step.
        
        """

        set_skforecast_warnings(suppress_warnings, action='ignore')

        (
            X,
            X_predict_dimension_names,
            *_
        ) = self._create_predict_inputs(
                steps        = steps,
                levels       = levels,
                last_window  = last_window,
                exog         = exog,
                check_inputs = check_inputs
            )

        X_predict = pd.DataFrame(
                        data    = X[0][0], 
                        columns = X_predict_dimension_names['X_autoreg'][2],
                        index   = X_predict_dimension_names['X_autoreg'][1] 
                    )
        
        exog_predict = None
        if self.exog_in_:
            exog_predict = pd.DataFrame(
                data    = X[1][0], 
                columns = X_predict_dimension_names['exog_pred'][2],
                index   = X_predict_dimension_names['exog_pred'][1]
            )
            # NOTE: not needed in this forecaster
            # categorical_features = any(
            #     not pd.api.types.is_numeric_dtype(dtype) or pd.api.types.is_bool_dtype(dtype) 
            #     for dtype in set(self.exog_dtypes_out_)
            # )
            # if categorical_features:
            #     X_predict = X_predict.astype(self.exog_dtypes_out_)
        
        if self.transformer_series is not None:
            warnings.warn(
                "The output matrix is in the transformed scale due to the "
                "inclusion of transformations in the Forecaster. "
                "As a result, any predictions generated using this matrix will also "
                "be in the transformed scale. Please refer to the documentation "
                "for more details: "
                "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html",
                DataTransformationWarning
            )
        
        set_skforecast_warnings(suppress_warnings, action='default')

        return X_predict, exog_predict

    def predict(
        self,
        steps: int | list[int] | None = None,
        levels: str | list[str] | None = None,
        last_window: pd.DataFrame | None = None,
        exog: pd.Series | pd.DataFrame | None = None,
        suppress_warnings: bool = False,
        check_inputs: bool = True
    ) -> pd.DataFrame:
        """
        Predict n steps ahead

        Parameters
        ----------
        steps : int, list, default None
            Predict n steps. The value of `steps` must be less than or equal to the 
            value of steps defined in the regressor architecture.
        
            - If `int`: Only steps within the range of 1 to int are predicted.
            - If `list`: List of ints. Only the steps contained in the list 
            are predicted.
            - If `None`: As many steps are predicted as defined in the regressor
            architecture.
        levels : str, list, default None
            Name(s) of the time series to be predicted. It must be included
            in `levels`, defined when initializing the forecaster. If `None`, all
            all series used during training will be available for prediction.
        last_window : pandas DataFrame, default `None`
            Series values used to create the predictors (lags) needed in the
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in `self.last_window_` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s.
        suppress_warnings : bool, default `False`
            If `True`, skforecast warnings will be suppressed during the fitting
            process. See skforecast.exceptions.warn_skforecast_categories for more
            information.
        check_inputs : bool, default True
            If `True`, the input is checked for possible warnings and errors 
            with the `check_predict_input` function. This argument is created 
            for internal use and is not recommended to be changed.

        Returns
        -------
        predictions : pandas DataFrame
            Predicted values.

        """

        set_skforecast_warnings(suppress_warnings, action="ignore")

        (
            X,
            _,
            steps,
            levels,
            prediction_index
        ) = self._create_predict_inputs(
                steps        = steps,
                levels       = levels,
                last_window  = last_window,
                exog         = exog,
                check_inputs = check_inputs
            )

        predictions = self.regressor.predict(
            X[0] if not self.exog_in_ else X, verbose=0
        )
        predictions = np.reshape(
            predictions, (predictions.shape[1], predictions.shape[2])
        )[np.array(steps) - 1]

        for i, level in enumerate(self.levels):
            # NOTE: The inverse transformation is applied only if the level
            # is included in the levels to predict.
            if level in levels:
                predictions[:, i] = transform_numpy(
                    array             = predictions[:, i],
                    transformer       = self.transformer_series_[level],
                    fit               = False,
                    inverse_transform = True
                )

        n_steps, n_levels = predictions.shape
        predictions = pd.DataFrame(
            {"level": np.tile(self.levels, n_steps), "pred": predictions.ravel()},
            index = np.repeat(prediction_index, n_levels),
        )
        predictions = predictions[predictions['level'].isin(levels)]

        set_skforecast_warnings(suppress_warnings, action="default")

        return predictions
    
    def _predict_interval_conformal(
        self,
        steps: int | list[int] | None = None,
        levels: str | list[str] | None = None,
        last_window: pd.Series | pd.DataFrame | None = None,
        exog: pd.Series | pd.DataFrame | None = None,
        nominal_coverage: float = 0.95,
        use_in_sample_residuals: bool = True
    ) -> pd.DataFrame:
        """
        Generate prediction intervals using the conformal prediction 
        split method [1]_.

        Parameters
        ----------
        steps : int, list, default None
            Predict n steps. The value of `steps` must be less than or equal to the 
            value of steps defined in the regressor architecture.
        
            - If `int`: Only steps within the range of 1 to int are predicted.
            - If `list`: List of ints. Only the steps contained in the list 
            are predicted.
            - If `None`: As many steps are predicted as defined in the regressor
            architecture.
        levels : str, list, default None
            Name(s) of the time series to be predicted. It must be included
            in `levels`, defined when initializing the forecaster. If `None`, all
            all series used during training will be available for prediction.
        last_window : pandas Series, pandas DataFrame, default None
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in` self.last_window_` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s.
        nominal_coverage : float, default 0.95
            Nominal coverage, also known as expected coverage, of the prediction
            intervals. Must be between 0 and 1.
        use_in_sample_residuals : bool, default True
            If `True`, residuals from the training data are used as proxy of
            prediction error to create predictions. 
            If `False`, out of sample residuals (calibration) are used. 
            Out-of-sample residuals must be precomputed using Forecaster's
            `set_out_sample_residuals()` method.

        Returns
        -------
        predictions : pandas DataFrame
            Values predicted by the forecaster and their estimated interval.

            - pred: predictions.
            - lower_bound: lower bound of the interval.
            - upper_bound: upper bound of the interval.

        References
        ----------
        .. [1] MAPIE - Model Agnostic Prediction Interval Estimator.
               https://mapie.readthedocs.io/en/stable/theoretical_description_regression.html#the-split-method

        """
        
        (
            X,
            _,
            steps,
            levels,
            prediction_index
        ) = self._create_predict_inputs(
                steps                   = steps,
                levels                  = levels,
                last_window             = last_window,
                exog                    = exog,
                predict_probabilistic   = True,
                use_in_sample_residuals = use_in_sample_residuals
            )

        if use_in_sample_residuals:
            residuals = self.in_sample_residuals_
        else:
            residuals = self.out_sample_residuals_

        predictions = self.regressor.predict(
            X[0] if not self.exog_in_ else X, verbose=0
        )
        predictions = np.reshape(
            predictions, (predictions.shape[1], predictions.shape[2])
        )[np.array(steps) - 1]
        
        n_steps = len(steps)
        n_levels = len(self.levels)
        correction_factor = np.full(
            shape=(n_steps, n_levels), fill_value=np.nan, order='C', dtype=float
        )
        for i, level in enumerate(self.levels):
            # NOTE: The correction factor is calculated only for the levels
            # included in the levels to predict.
            if level in levels:
                correction_factor[:, i] = np.quantile(
                    np.abs(residuals[level]), nominal_coverage
                )
            else:
                correction_factor[:, i] = 0.

        lower_bound = predictions - correction_factor
        upper_bound = predictions + correction_factor

        # NOTE: Create a 3D array with shape (n_levels, intervals, steps)
        predictions = np.array([predictions, lower_bound, upper_bound]).swapaxes(0, 2)

        for i, level in enumerate(self.levels):
            # NOTE: The inverse transformation is applied only if the level
            # is included in the levels to predict.
            if level in levels:
                transformer_level = self.transformer_series_[level]
                if transformer_level is not None:
                    predictions[i, :, :] = np.apply_along_axis(
                        func1d            = transform_numpy,
                        axis              = 0,
                        arr               = predictions[i, :, :],
                        transformer       = transformer_level,
                        fit               = False,
                        inverse_transform = True
                    )
        
        predictions = pd.DataFrame(
                          data    = predictions.swapaxes(0, 1).reshape(-1, 3),
                          index   = np.repeat(prediction_index, len(self.levels)),
                          columns = ["pred", "lower_bound", "upper_bound"]
                      )
        predictions.insert(0, 'level', np.tile(self.levels, n_steps))
        predictions = predictions[predictions['level'].isin(levels)]

        return predictions

    def predict_interval(
        self,
        steps: int | list[int] | None = None,
        levels: str | list[str] | None = None,
        last_window: pd.DataFrame | None = None,
        exog: pd.Series | pd.DataFrame | None = None,
        method: str = 'conformal',
        interval: float | list[float] | tuple[float] = [5, 95],
        use_in_sample_residuals: bool = True,
        suppress_warnings: bool = False,
        n_boot: Any = None,
        use_binned_residuals: Any = None,
        random_state: Any = None,
    ) -> pd.DataFrame:
        """
        Predict n steps ahead and estimate prediction intervals using conformal 
        prediction method. Refer to the References section for additional details.
        
        Parameters
        ----------
        steps : int, list, default None
            Predict n steps. The value of `steps` must be less than or equal to the 
            value of steps defined in the regressor architecture.
        
            - If `int`: Only steps within the range of 1 to int are predicted.
            - If `list`: List of ints. Only the steps contained in the list 
            are predicted.
            - If `None`: As many steps are predicted as defined in the regressor
            architecture.
        levels : str, list, default None
            Name(s) of the time series to be predicted. It must be included
            in `levels`, defined when initializing the forecaster. If `None`, all
            all series used during training will be available for prediction.
        last_window : pandas DataFrame, default None
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in `self.last_window_` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, dict, default None
            Exogenous variable/s included as predictor/s.
        method : str, default 'conformal'
            Employs the conformal prediction split method for interval estimation [1]_.
        interval : float, list, tuple, default [5, 95]
            Confidence level of the prediction interval. Interpretation depends 
            on the method used:
            
            - If `float`, represents the nominal (expected) coverage (between 0 
            and 1). For instance, `interval=0.95` corresponds to `[2.5, 97.5]` 
            percentiles.
            - If `list` or `tuple`, defines the exact percentiles to compute, which 
            must be between 0 and 100 inclusive. For example, interval 
            of 95% should be as `interval = [2.5, 97.5]`.
            - When using `method='conformal'`, the interval must be a float or 
            a list/tuple defining a symmetric interval.
        use_in_sample_residuals : bool, default True
            If `True`, residuals from the training data are used as proxy of
            prediction error to create predictions. 
            If `False`, out of sample residuals (calibration) are used. 
            Out-of-sample residuals must be precomputed using Forecaster's
            `set_out_sample_residuals()` method.
        suppress_warnings : bool, default False
            If `True`, skforecast warnings will be suppressed during the prediction 
            process. See skforecast.exceptions.warn_skforecast_categories for more
            information.
        n_boot : Ignored
            Not used, present here for API consistency by convention.
        use_binned_residuals : Ignored
            Not used, present here for API consistency by convention.
        random_state : Ignored
            Not used, present here for API consistency by convention.

        Returns
        -------
        predictions : pandas DataFrame
            Long-format DataFrame with the predictions and the lower and upper
            bounds of the estimated interval. The columns are `level`, `pred`,
            `lower_bound`, `upper_bound`.

        References
        ----------        
        .. [1] MAPIE - Model Agnostic Prediction Interval Estimator.
               https://mapie.readthedocs.io/en/stable/theoretical_description_regression.html#the-split-method
    
        """

        set_skforecast_warnings(suppress_warnings, action='ignore')

        if method == "conformal":

            if isinstance(interval, (list, tuple)):
                check_interval(interval=interval, ensure_symmetric_intervals=True)
                nominal_coverage = (interval[1] - interval[0]) / 100
            else:
                check_interval(alpha=interval, alpha_literal='interval')
                nominal_coverage = interval
            
            predictions = self._predict_interval_conformal(
                              steps                   = steps,
                              levels                  = levels,
                              last_window             = last_window,
                              exog                    = exog,
                              nominal_coverage        = nominal_coverage,
                              use_in_sample_residuals = use_in_sample_residuals
                          )
        else:
            raise ValueError(
                f"Invalid `method` '{method}'. Only 'conformal' is available."
            )
        
        set_skforecast_warnings(suppress_warnings, action='default')

        return predictions

    def plot_history(
        self,
        ax: matplotlib.axes.Axes = None,
        exclude_first_iteration: bool = False,
        **fig_kw,
    ) -> matplotlib.figure.Figure:
        """
        Plots the training and validation loss curves from the given history object stored
        in the ForecasterRnn.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, default `None`
            Pre-existing ax for the plot. Otherwise, call matplotlib.pyplot.subplots()
            internally.
        exclude_first_iteration : bool, default `False`
            Whether to exclude the first epoch from the plot.
        fig_kw : dict
            Other keyword arguments are passed to matplotlib.pyplot.subplots().

        Returns
        -------
        fig: matplotlib.figure.Figure
            Matplotlib Figure.
        
        """

        if ax is None:
            fig, ax = plt.subplots(1, 1, **fig_kw)
        else:
            fig = ax.get_figure()

        # Setting up the plot style
        if self.history_ is None:
            raise ValueError("ForecasterRnn has not been fitted yet.")

        # Determine the range of epochs to plot, excluding the first one if specified
        epoch_range = range(1, len(self.history_["loss"]) + 1)
        if exclude_first_iteration:
            epoch_range = range(2, len(self.history_["loss"]) + 1)

        # Plotting training loss
        ax.plot(
            epoch_range,
            self.history_["loss"][
                exclude_first_iteration:
            ],  # Skip first element if specified
            color="b",
            label="Training Loss",
        )

        # Plotting validation loss
        if "val_loss" in self.history_:
            ax.plot(
                epoch_range,
                self.history_["val_loss"][
                    exclude_first_iteration:
                ],  # Skip first element if specified
                color="r",
                label="Validation Loss",
            )

        # Labeling the axes and adding a title
        ax.set_xlabel("Epochs")
        ax.set_ylabel("Loss")
        ax.set_title("Training and Validation Loss")

        # Adding a legend
        ax.legend()

        # Displaying grid for better readability
        ax.grid(True, linestyle="--", alpha=0.7)

        # Setting x-axis ticks to integers only
        ax.set_xticks(epoch_range)

        return fig

    # TODO create testing
    def set_params(self, params: dict) -> None:  
        """
        Set new values to the parameters of the scikit-learn model stored in the
        forecaster. It is important to note that all models share the same
        configuration of parameters and hyperparameters.

        Parameters
        ----------
        params : dict
            Parameters values.

        Returns
        -------
        None

        """

        self.regressor = clone(self.regressor)
        self.regressor.reset_states()
        self.regressor.compile(**params)

    def set_fit_kwargs(self, fit_kwargs: dict) -> None:
        """
        Set new values for the additional keyword arguments passed to the `fit`
        method of the regressor.

        Parameters
        ----------
        fit_kwargs : dict
            Dict of the form {"argument": new_value}.

        Returns
        -------
        None

        """

        self.fit_kwargs = check_select_fit_kwargs(self.regressor, fit_kwargs=fit_kwargs)

    def set_lags(self, lags: Any) -> None:  # pragma: no cover
        """
        Not used, present here for API consistency by convention.

        Returns
        -------
        None

        """

        pass

    def set_in_sample_residuals(
        self,
        series: pd.DataFrame,
        exog: pd.Series | pd.DataFrame = None,
        random_state: int = 123,
        suppress_warnings: bool = False
    ) -> None:
        """
        Set in-sample residuals in case they were not calculated during the
        training process. 
        
        In-sample residuals are calculated as the difference between the true 
        values and the predictions made by the forecaster using the training 
        data. The following internal attributes are updated:

        + `in_sample_residuals_`: Dictionary containing a numpy ndarray with the
        residuals for each series in the form `{series: residuals}`.

        A total of 10_000 residuals are stored in the attribute `in_sample_residuals_`.
        If the number of residuals is greater than 10_000, a random sample of
        10_000 residuals is stored. The number of residuals stored per bin is
        limited to `10_000 // self.binner.n_bins_`.
        
        Parameters
        ----------
        series : pandas DataFrame
            Training time series.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s.
        random_state : int, default 123
            Sets a seed to the random sampling for reproducible output.
        suppress_warnings : bool, default False
            If `True`, skforecast warnings will be suppressed during the sampling 
            process. See skforecast.exceptions.warn_skforecast_categories for more
            information.

        Returns
        -------
        None

        """

        set_skforecast_warnings(suppress_warnings, action='ignore')

        if not self.is_fitted:
            raise NotFittedError(
                "This forecaster is not fitted yet. Call `fit` with appropriate "
                "arguments before using `set_in_sample_residuals()`."
            )

        if not isinstance(series, pd.DataFrame):
            raise TypeError(
                f"`series` must be a pandas DataFrame. Got {type(series)}."
            )
        
        series_index_range = check_extract_values_and_index(
            data=series, data_label='`series`', return_values=False
        )[1][[0, -1]]
        if not series_index_range.equals(self.training_range_):
            raise IndexError(
                f"The index range of `series` does not match the range "
                f"used during training. Please ensure the index is aligned "
                f"with the training data.\n"
                f"    Expected : {self.training_range_}\n"
                f"    Received : {series_index_range}"
            )
            
        (
            X_train,
            exog_train,
            y_train,
            dimension_names,
            *_
        ) = self._create_train_X_y(series=series, exog=exog)
        
        if exog is not None:
            X_train_features_names_out_ = dimension_names["X_train"][1] + dimension_names["exog_train"][2]
        else:
            X_train_features_names_out_ = dimension_names["X_train"][1]
        
        if not X_train_features_names_out_ == self.X_train_features_names_out_:
            raise ValueError(
                f"Feature mismatch detected after matrix creation. The features "
                f"generated from the provided data do not match those used during "
                f"the training process. To correctly set in-sample residuals, "
                f"ensure that the same data and preprocessing steps are applied.\n"
                f"    Expected output : {self.X_train_features_names_out_}\n"
                f"    Current output  : {X_train_features_names_out_}"
            )
        
        # TODO: Include binning in the forecaster
        self.in_sample_residuals_ = {}
        residuals = y_train - self.regressor.predict(
            x=X_train if exog_train is None else [X_train, exog_train], verbose=0
        )
        residuals = np.concatenate(
            [residuals[:, i, :] for i, step in enumerate(self.steps)]
        )

        rng = np.random.default_rng(seed=random_state)
        for i, level in enumerate(self.levels):
            residuals_level = residuals[:, i]
            if len(residuals_level) > 10_000:
                residuals_level = residuals_level[
                    rng.integers(low=0, high=len(residuals_level), size=10_000)
                ]
            self.in_sample_residuals_[level] = residuals_level

        set_skforecast_warnings(suppress_warnings, action='default')

    def set_out_sample_residuals(
        self,
        y_true: dict[str, np.ndarray | pd.Series],
        y_pred: dict[str, np.ndarray | pd.Series],
        append: bool = False,
        random_state: int = 123
    ) -> None:
        """
        Set new values to the attribute `out_sample_residuals_`. Out of sample
        residuals are meant to be calculated using observations that did not
        participate in the training process. `y_true` and `y_pred` are expected
        to be in the original scale of the time series. Residuals are calculated
        as `y_true` - `y_pred`, after applying the necessary transformations and
        differentiations if the forecaster includes them (`self.transformer_series`
        and `self.differentiation`).

        A total of 10_000 residuals are stored in the attribute `out_sample_residuals_`.
        If the number of residuals is greater than 10_000, a random sample of
        10_000 residuals is stored.
        
        Parameters
        ----------
        y_true : dict
            Dictionary of numpy ndarrays or pandas Series with the true values of
            the time series for each series in the form {series: y_true}.
        y_pred : dict
            Dictionary of numpy ndarrays or pandas Series with the predicted values
            of the time series for each series in the form {series: y_pred}.
        append : bool, default False
            If `True`, new residuals are added to the once already stored in the
            attribute `out_sample_residuals_`. If after appending the new residuals,
            the limit of 10_000 samples is exceeded, a random sample of 10_000 is
            kept.
        random_state : int, default 123
            Sets a seed to the random sampling for reproducible output.

        Returns
        -------
        None

        """

        if not self.is_fitted:
            raise NotFittedError(
                "This forecaster is not fitted yet. Call `fit` with appropriate "
                "arguments before using `set_out_sample_residuals()`."
            )

        if not isinstance(y_true, dict):
            raise TypeError(
                f"`y_true` must be a dictionary of numpy ndarrays or pandas Series. "
                f"Got {type(y_true)}."
            )
  
        if not isinstance(y_pred, dict):
            raise TypeError(
                f"`y_pred` must be a dictionary of numpy ndarrays or pandas Series. "
                f"Got {type(y_pred)}."
            )
        
        if not set(y_true.keys()) == set(y_pred.keys()):
            raise ValueError(
                f"`y_true` and `y_pred` must have the same keys. "
                f"Got {set(y_true.keys())} and {set(y_pred.keys())}."
            )
        
        for k in y_true.keys():
            if not isinstance(y_true[k], (np.ndarray, pd.Series)):
                raise TypeError(
                    f"Values of `y_true` must be numpy ndarrays or pandas Series. "
                    f"Got {type(y_true[k])} for series {k}."
                )
            if not isinstance(y_pred[k], (np.ndarray, pd.Series)):
                raise TypeError(
                    f"Values of `y_pred` must be numpy ndarrays or pandas Series. "
                    f"Got {type(y_pred[k])} for series {k}."
                )
            if len(y_true[k]) != len(y_pred[k]):
                raise ValueError(
                    f"`y_true` and `y_pred` must have the same length. "
                    f"Got {len(y_true[k])} and {len(y_pred[k])} for series {k}."
                )
            if isinstance(y_true[k], pd.Series) and isinstance(y_pred[k], pd.Series):
                if not y_true[k].index.equals(y_pred[k].index):
                    raise ValueError(
                        f"When containing pandas Series, elements in `y_true` and "
                        f"`y_pred` must have the same index. Error in series {k}."
                    )
        
        series_to_update = set(y_pred.keys()).intersection(set(self.levels))
        if not series_to_update:
            raise ValueError(
                f"Provided keys in `y_pred` and `y_true` do not match any of the "
                f"target time series in the forecaster, {self.levels}. Residuals "
                f"cannot be updated."
            )
        
        if self.out_sample_residuals_ is None:
            self.out_sample_residuals_ = {level: None for level in self.levels}
        
        rng = np.random.default_rng(seed=random_state)
        for level in series_to_update:

            y_true_level = deepcopy(y_true[level])
            y_pred_level = deepcopy(y_pred[level])
            if not isinstance(y_true_level, np.ndarray):
                y_true_level = y_true_level.to_numpy()
            if not isinstance(y_pred_level, np.ndarray):
                y_pred_level = y_pred_level.to_numpy()

            if self.transformer_series:
                y_true_level = transform_numpy(
                                   array             = y_true_level,
                                   transformer       = self.transformer_series_[level],
                                   fit               = False,
                                   inverse_transform = False
                               )
                y_pred_level = transform_numpy(
                                   array             = y_pred_level,
                                   transformer       = self.transformer_series_[level],
                                   fit               = False,
                                   inverse_transform = False
                               )

            data = pd.DataFrame(
                {'prediction': y_pred_level, 'residuals': y_true_level - y_pred_level}
            ).dropna()
            residuals = data['residuals'].to_numpy()

            out_sample_residuals = self.out_sample_residuals_.get(level, np.array([]))
            out_sample_residuals = (
                np.array([]) 
                if out_sample_residuals is None
                else out_sample_residuals
            )
            if append:
                out_sample_residuals = np.concatenate([out_sample_residuals, residuals])
            else:
                out_sample_residuals = residuals

            if len(out_sample_residuals) > 10_000:
                out_sample_residuals = rng.choice(
                    a       = out_sample_residuals, 
                    size    = 10_000, 
                    replace = False
                )

            self.out_sample_residuals_[level] = out_sample_residuals
